# ruff: noqa: N802, N803, N806

import numpy as np
import pytest
from scipy.spatial.transform import Rotation as ScipyRotation

import archimedes as arc
from archimedes.spatial import (
    EulerAngles,
    Quaternion,
    quaternion_inverse,
    quaternion_multiply,
    quaternion_to_dcm,
    quaternion_to_euler,
)

np.random.seed(0)


def random_quat(wrapper=False):
    """Generate a random rotation"""
    rand_quat = np.random.randn(4)
    rand_quat /= np.linalg.norm(rand_quat)
    if wrapper:
        rand_quat = Quaternion(rand_quat)
    return rand_quat


class TestQuaternionLowLevel:
    def test_multiplication(self):
        R1, R2 = random_quat(), random_quat()
        q1 = quaternion_multiply(R1, R2)
        R1_scipy = ScipyRotation.from_quat(R1, scalar_first=True)
        R2_scipy = ScipyRotation.from_quat(R2, scalar_first=True)
        q2 = (R1_scipy * R2_scipy).as_quat(scalar_first=True)
        assert np.allclose(q1, q2)

    def test_inverse(self):
        q = random_quat()
        q_inv = quaternion_inverse(q)
        q_inv_scipy = (
            ScipyRotation.from_quat(q, scalar_first=True)
            .inv()
            .as_quat(scalar_first=True)
        )
        assert np.allclose(q_inv, q_inv_scipy)

    def test_to_dcm(self):
        quat = random_quat()
        R1 = quaternion_to_dcm(quat)
        R2 = ScipyRotation.from_quat(quat, scalar_first=True).as_matrix()
        # NOTE: SciPy's rotation uses an "active" convention (rotating vectors
        # in a single frame), whereas our DCMs are "passive" (rotating between
        # frames).  Thus the matrices are transposes of each other.
        np.testing.assert_allclose(R1, R2.T)

    def test_to_euler(self):
        q = random_quat()
        for seq in ("XYZ", "xyz", "ZYX", "zyx", "XYX", "zxz"):
            euler1 = quaternion_to_euler(q, seq=seq)
            euler2 = ScipyRotation.from_quat(q, scalar_first=True).as_euler(seq)
            np.testing.assert_allclose(euler1, euler2)


class TestQuaternionWrapper:
    def test_ops(self):
        q = Quaternion.identity()
        assert len(q) == 4

        # __getitem__x
        assert q[0] == 1.0

        # __iter__
        w, x, y, z = q
        assert np.allclose(np.array([w, x, y, z]), np.array([1.0, 0.0, 0.0, 0.0]))

        assert q.as_quat() is q
        np.testing.assert_allclose(Quaternion.from_quat(q).array, q.array)

    def test_identity(self):
        q = Quaternion.identity()
        assert np.allclose(q.as_matrix(), np.eye(3))

    def test_rotate(self):
        q = random_quat(wrapper=True)
        v = np.array([0.1, 0.2, 0.3])
        w = q.as_matrix() @ v

        R_scipy = ScipyRotation.from_quat(q.array, scalar_first=True)
        w_scipy = R_scipy.apply(v, inverse=True)

        assert np.allclose(w, w_scipy)

    def test_multiplication(self):
        q1, q2 = random_quat(True), random_quat(True)
        q3 = q1 * q2
        R1_scipy = ScipyRotation.from_quat(q1.array, scalar_first=True)
        R2_scipy = ScipyRotation.from_quat(q2.array, scalar_first=True)
        q3_scipy = (R1_scipy * R2_scipy).as_quat(scalar_first=True)
        assert np.allclose(q3.array, q3_scipy)

    def test_composition_associativity(self):
        R1, R2, R3 = [random_quat(True) for _ in range(3)]
        q1 = (R1 * R2) * R3
        q2 = R1 * (R2 * R3)
        assert np.allclose(q1.array, q2.array)

    def test_inverse(self):
        q = random_quat(True)
        q0 = Quaternion.identity()
        q1 = q * q.inv()
        q2 = q.inv() * q
        assert np.allclose(q1.array, q0.array)
        assert np.allclose(q2.array, q0.array)

    def test_kinematics(self):
        q = random_quat(wrapper=True)
        w_B = np.array([0.01, -0.02, 0.03])
        q_t1 = q.kinematics(w_B, baumgarte=1.0)
        q_t2 = 0.5 * quaternion_multiply(q.array, np.array([0, *w_B]))
        assert np.allclose(q_t1.array, q_t2, atol=1e-6)

    def _quat_roundtrip(self, euler_orig, seq, debug=False):
        q = Quaternion.from_euler(euler_orig, seq)
        assert len(q) == 4
        euler2 = q.as_euler(seq)

        if debug:
            R2 = ScipyRotation.from_euler(seq, euler_orig)
            print(f"quat:       {q.array}")
            print(f"SciPy quat: {R2.as_quat(scalar_first=True)}")

            print(f"euler:       {euler2}")
            print(f"SciPy euler: {R2.as_euler(seq)}")

        assert np.allclose(euler2.array, euler_orig)

    def _dcm_roundtrip(self, euler_orig, seq, debug=False):
        # Euler -> matrix -> quat -> matrix -> euler
        q1 = Quaternion.from_euler(euler_orig, seq)
        q2 = Quaternion.from_matrix(q1.as_matrix())
        euler2 = q2.as_euler(seq)

        if debug:
            R1_scipy = ScipyRotation.from_euler(seq, euler_orig)
            R2_scipy = ScipyRotation.from_matrix(q1.as_matrix())
            print(f"quat:       {q1.array}")
            print(f"SciPy quat: {R1_scipy.as_quat(scalar_first=True)}")

            print(f"euler:       {euler2}")
            print(f"SciPy euler: {R2_scipy.as_euler(seq)}")

        assert np.allclose(euler2.array, euler_orig)

    @pytest.mark.parametrize(
        "seq",
        [
            "xyz",  # Standard roll-pitch-yaw
            "zyx",  # Standard yaw-pitch-roll
            "zxz",  # Symmetric sequence
            "ZYX",  # Intrinsic sequence
            "XZX",  # Symmetric intrinsic sequence
        ],
    )
    def test_roundtrip(self, seq):
        euler_orig = np.array([0.1, 0.2, 0.3])
        self._quat_roundtrip(euler_orig, seq)
        self._dcm_roundtrip(euler_orig, seq)

    @pytest.mark.parametrize(
        "angles",
        [
            [0, 0, 0],  # Identity
            [np.pi / 2, 0, 0],  # 90° roll
            [0.1, 0.2, 0.3],  # Small angles
            [np.pi - 0.1, 0.1, 0.1],  # Near-singularity
        ],
    )
    def test_with_scipy(self, angles):
        seq = "xyz"

        # Both libraries should give same results
        R_scipy = ScipyRotation.from_euler(seq, angles)
        q = Quaternion.from_euler(angles, seq)

        assert np.allclose(q.array, R_scipy.as_quat(scalar_first=True))

    def test_compile(self):
        @arc.compile
        def rotate_vector(q, v):
            R = q.as_matrix()
            return R @ v

        q = Quaternion.from_euler([0.1, 0.2, 0.3], "xyz")
        v = np.array([1, 2, 3])

        result = rotate_vector(q, v)
        assert result.shape == (3,)
        assert np.allclose(result, q.as_matrix() @ v)

    def test_tree_ops(self):
        q = Quaternion.from_euler([0.1, 0.2, 0.3], "xyz")
        flat, unflatten = arc.tree.ravel(q)
        q_restored = unflatten(flat)

        R1 = q.as_matrix()
        R2 = q_restored.as_matrix()
        assert np.allclose(R1, R2)

    def test_errors(self):
        # Invalid output sequence
        with pytest.raises(ValueError, match="Expected `seq` to be a string"):
            Quaternion.identity().as_euler("xz")

        # Pass sequence and Euler angles
        with pytest.raises(ValueError, match="If `euler` is an EulerAngles"):
            angles = EulerAngles([0.1, 0.2, 0.3], seq="xyz")
            Quaternion.from_euler(angles, seq="zyx")

        # Invalid quat shape
        with pytest.raises(ValueError, match="Quaternion must have shape"):
            Quaternion(np.array([1.0, 2.0, 3.0]))

        # Invalid matrix shape
        with pytest.raises(ValueError, match="Rotation matrix must be 3x3"):
            Quaternion.from_matrix(np.eye(4))



import numpy as np
import pytest

from archimedes._core import SymbolicArray, array, compile, sym
from archimedes.optimize import minimize

METHODS = ("ipopt", "sqpmethod", "BFGS")
BOUNDED_METHODS = ("ipopt", "sqpmethod", "L-BFGS-B")


class TestMinimize:
    @pytest.mark.parametrize("method", METHODS)
    def test_minimize(self, method):
        # Basic functionality test
        def f(x):
            return x**2

        result = minimize(f, x0=1.0, method=method)
        assert np.allclose(result.x, 0.0)

    def test_minimize_with_param(self):
        # Test with a parameter
        def f(x, a=1.0):
            return a * x**2

        result = minimize(f, x0=1.0, args=(2.0,))
        assert np.allclose(result.x, 0.0)

    @pytest.mark.parametrize("method", BOUNDED_METHODS)
    def test_minimize_with_bounds(self, method):
        def f(x):
            return x**2

        result = minimize(f, x0=1.0, bounds=[-2.0, 2.0], method=method)
        assert np.allclose(result.x, 0.0)

    @pytest.mark.parametrize("method", ("ipopt", "sqpmethod"))
    def test_minimize_constrained(self, method):
        # Test with function from the CasADi docs, using additional parameters
        def f(x, a, b):
            return x[0] ** 2 + a * x[2] ** 2

        def g(x, a, b):
            return x[2] + b * (1 - x[0]) ** 2 - x[1]

        x0 = np.random.randn(3)
        args = (100.0, 1.0)
        result = minimize(f, constr=g, x0=x0, args=args, method=method)
        assert np.allclose(result.x, [0.0, 1.0, 0.0])

        # Test with bounds
        result_bounded = minimize(
            f,
            constr=g,
            x0=x0,
            args=args,
            bounds=(np.full(3, -10), np.full(3, 10)),
            method=method,
        )
        assert np.allclose(result.x, result_bounded.x)

        # Test symbolic evaluation
        x0 = sym("x", shape=(3,), kind="MX")
        result = minimize(f, x0=x0, args=args, constr=g)
        assert isinstance(result.x, SymbolicArray)

    def test_minimize_rosenbrock(self):
        # Test the Rosenbrock function
        def f(x, a):
            return a * (x[1] - x[0] ** 2) ** 2 + (1 - x[0]) ** 2

        result = minimize(f, x0=[-1.0, 1.0], static_argnames=("a",), args=(100.0,))
        assert np.allclose(result.x, [1.0, 1.0])

    def test_minimize_tree(self):
        def f(params):
            x, y = params["x"], params["y"]
            return 100 * (y - x**2) ** 2 + (1 - x) ** 2

        def g(params):
            x, y = params["x"], params["y"]
            return x + y - 1.5  # x + y >= 1.5

        # Tree-structured initial guess
        x0 = {"x": 2.0, "y": 1.0}

        # Solve with inequality constraint
        result = minimize(f, x0, constr=g, constr_bounds=(0.0, np.inf))

        assert np.allclose(result.x["x"], 1.0, atol=1e-3)
        assert np.allclose(result.x["y"], 1.0, atol=1e-3)

    def test_minimize_rosenbrock_constrained(self):
        # https://en.wikipedia.org/wiki/Test_functions_for_optimization
        #
        # This has a local minimum at (0, 0) and a global minimum at (1, 1)

        def f(x):
            return 100 * (x[1] - x[0] ** 2) ** 2 + (1 - x[0]) ** 2

        def g(x):
            g1 = (x[0] - 1) ** 3 - x[1] + 1
            g2 = x[0] + x[1] - 2
            return array([g1, g2])

        result = minimize(f, constr=g, x0=[2.0, 0.0], constr_bounds=(-np.inf, 0))
        assert np.allclose(result.x, [1.0, 1.0], atol=1e-3)

    def test_error_handling(self):
        # Inconsistent arguments
        def f(x):
            return x**2

        def g(x, y):
            return x + y

        with pytest.raises(ValueError, match=r".*must have the same number"):
            minimize(f, constr=g, x0=1.0)

        # Unsupported method
        with pytest.raises(ValueError):
            minimize(f, x0=1.0, method="unsupported_method")

        # Inconsistent static arguments
        def f(a, x):
            return a * x**2

        def g(a, x):
            return a * x

        f = compile(f, static_argnames=["a"])
        g = compile(g)

        with pytest.raises(ValueError, match=r".*must have the same number"):
            minimize(f, constr=g, x0=1.0)



"""High-level wrappers for attitude representations."""

# ruff: noqa: N803, N806
from __future__ import annotations

from typing import Protocol

import numpy as np

from archimedes import field, struct

from ._euler import _check_angles, _check_seq, euler_kinematics, euler_to_dcm
from ._quaternion import (
    dcm_to_quaternion,
    euler_to_quaternion,
    quaternion_inverse,
    quaternion_kinematics,
    quaternion_multiply,
    quaternion_to_dcm,
    quaternion_to_euler,
)

__all__ = [
    "Attitude",
    "EulerAngles",
    "Quaternion",
]


class Attitude(Protocol):
    """Protocol for attitude representations.

    This protocol defines the interface that all attitude representation classes
    must implement. An attitude representation class encapsulates the orientation
    of a rigid body in 3D space and provides methods for common operations such
    as conversion to/from other representations and kinematic models.
    """

    def as_matrix(self) -> np.ndarray:
        """Convert the attitude to a direction cosine matrix (DCM).

        If the attitude represents the orientation of a body A relative to a frame B,
        then this method returns the matrix R_BA that transforms vectors from
        frame A to frame B.  Specifically, for a vector v_A expressed in frame A,
        the corresponding vector in frame B is given by ``v_B = R_BA @ v_A``.

        The inverse transformation can be obtained by transposing this matrix:
        ``R_AB = R_BA.T``.

        Returns
        -------
            A 3x3 numpy array representing the DCM.
        """

    def inv(self) -> Attitude:
        """Compute the inverse of the rotation corresponding to the attitude.

        Returns
        -------
        Attitude
            A new attitude instance representing the inverse rotation.
        """

    def kinematics(self, w_B: np.ndarray) -> Attitude:
        """Compute the time derivative of the attitude given angular velocity.

        **CAUTION**: This method returns the time derivative of the attitude,
        which is represented with the same data structure for consistency with
        ODE solving - but this return is not itself a valid rotation representation
        until integrated in time. Hence, the output of ``kinematics`` should never
        be converted to a different attitude representation or rotation matrix.

        Parameters
        ----------
        w_B : np.ndarray
            Angular velocity vector expressed in the body frame B.

        Returns
        -------
        Attitude
            The time derivative of the attitude representation.
        """


@struct
class EulerAngles:
    """Euler angle representation of a rotation in 3 dimensions

    Parameters
    ----------
    angles : array_like
        Euler angles in radians
    seq : str, optional
        Sequence of axes for Euler angles (up to length 3).  Each character must be one
        of 'x', 'y', 'z' (extrinsic) or 'X', 'Y', 'Z' (intrinsic).  Default is 'xyz'.

    Attributes
    ----------
    array : np.ndarray
        Underlying array of Euler angles.
    seq : str
        Sequence of axes for Euler angles.

    Examples
    --------
    >>> from archimedes.spatial import EulerAngles
    >>> import numpy as np

    Consider a right-handed rotation of 90 degrees about the z-axis. This
    corresponds to a single yaw rotation:

    >>> euler = EulerAngles(np.deg2rad(90), 'z')

    This can be converted to other representations:

    >>> euler.as_matrix()
    array([[ 6.123234e-17,  1.000000e+00,  0.000000e+00],
       [-1.000000e+00,  6.123234e-17,  0.000000e+00],
       [ 0.000000e+00,  0.000000e+00,  1.000000e+00]])
    >>> np.rad2deg(euler.as_euler('xyx'))  # Roll-pitch-roll sequence
    array([-90.,  90.,  90.])
    >>> euler.as_quat()
    Quaternion([0.70710678 0.         0.         0.70710678])

    The associated rotation matrix can be used to change coordinate systems. If the
    Euler sequence represents the orientation of a frame B relative to a frame A, then
    the rotation matrix transforms vectors from frame A to frame B:

    >>> v_A = np.array([1, 0, 0])  # Vector in frame A
    >>> R_BA = euler.as_matrix()
    >>> R_BA @ v_A  # Vector in frame B
    [6.12323e-17, -1, 0]

    The ``kinematics`` method can be used to compute the time derivative of the
    Euler angles given the angular velocity in the body frame. Note that Euler
    kinematics are currently only supported for the "xyz" sequence (standard
    roll-pitch-yaw):

    >>> w_B = np.array([0, 0, np.pi/2])  # 90 deg/s about z-axis
    >>> rpy = EulerAngles([0.1, 0.2, 0.3], "xyz")
    >>> rpy.kinematics(w_B)
    EulerAngles([ 0.31682542 -0.15681796  1.59473746], seq='xyz')

    Be careful with the kinematics output; this is expressed as an ``EulerAngles``
    instance for consistency with ODE solvers but represents rotation _rates_.
    Trying to apply this output as a rotation or convert to a DCM will not produce
    meaningful results.

    See Also
    --------
    Quaternion : Quaternion representation of rotation in 3D
    RigidBody : Rigid body dynamics supporting ``EulerAngles`` attitude representation
    euler_to_dcm : Directly calculate rotation matrix from roll-pitch-yaw angles
    euler_kinematics : Transform roll-pitch-yaw rates to body-frame angular velocity
    """

    array: np.ndarray
    seq: str = field(static=True, default="xyz")  # type: ignore

    def __post_init__(self):
        angles = self.array
        seq = self.seq
        # Internal implementation: for some analysis (e.g. vmap), the
        # struct might be initialized with an int instead of an array
        # for determining axis indices.  In that case we don't want to
        # do explicit shape validation
        if not isinstance(angles, int):
            _check_seq(seq)
            # Also for vmapping: angles might be passed as a 2D array
            # This solution isn't ideal, since it could introduce user errors
            # by inadvertently passing a 2D array.  It might be better to
            # check the angles when methods are called instead.
            if not getattr(angles, "ndim", None) == 2:
                angles = _check_angles(angles, seq)

        # NOTE: object.__setattr__ is used instead of directly setting because
        # this is a frozen dataclass
        object.__setattr__(self, "array", angles)

    # === Methods for implementing Attitude protocol ===

    def as_matrix(self) -> np.ndarray:
        """Convert the Euler angles to a direction cosine matrix (DCM).

        If the Euler angles represent the orientation of a body A relative to a frame
        B, then this method returns the matrix R_BA that transforms vectors from
        frame A to frame B.  Specifically, for a vector v_A expressed in frame A,
        the corresponding vector in frame B is given by ``v_B = R_BA @ v_A``.

        The inverse transformation can be obtained by transposing this matrix:
        ``R_AB = R_BA.T``.

        Returns
        -------
            A 3x3 numpy array representing the DCM.
        """
        return euler_to_dcm(self.array, self.seq)

    def inv(self) -> EulerAngles:
        """Return the inverse (conjugate) of this Euler angle rotation.

        Returns
        -------
        EulerAngles
            A new EulerAngles instance representing the inverse rotation.
        """
        angles = -self.array[::-1]
        seq = self.seq[::-1]
        return EulerAngles(angles, seq=seq)

    def kinematics(self, w_B: np.ndarray) -> EulerAngles:
        """Compute the time derivative of the Euler angles given angular velocity.

        **CAUTION**: This method returns the time derivative of the attitude,
        which is represented with the same data structure for consistency with
        ODE solving - but this return is not itself a valid rotation representation
        until integrated in time. Hence, the output of ``kinematics`` should never
        be converted to a different attitude representation or rotation matrix.

        Parameters
        ----------
        w_B : np.ndarray
            Angular velocity vector expressed in the body frame B.

        Returns
        -------
        EulerAngles
            Time derivative of the Euler angles.
        """
        if self.seq != "xyz":
            raise NotImplementedError(
                "Euler angle kinematics currently only implemented for 'xyz' sequence"
            )
        H = euler_kinematics(self.array, inverse=False)
        return EulerAngles(H @ w_B)

    # === Other methods ===

    def __repr__(self) -> str:
        return f"EulerAngles({self.array}, seq='{self.seq}')"

    def __len__(self) -> int:
        """Return the number of Euler angles (length of sequence)."""
        return len(self.seq)

    def __getitem__(self, index):
        return self.array[index]

    def __iter__(self):
        return iter(self.array)

    @classmethod
    def from_quat(cls, quat: Quaternion | np.ndarray, seq: str = "xyz") -> EulerAngles:
        """Create EulerAngles from a Quaternion.

        Parameters
        ----------
        quat : Quaternion or array_like
            Quaternion representing the rotation.

        Returns
        -------
        EulerAngles
            New EulerAngles instance representing the same rotation.
        """
        if isinstance(quat, Quaternion):
            quat = quat.array
        angles = quaternion_to_euler(quat, seq=seq)
        return cls(angles, seq=seq)

    def as_quat(self) -> Quaternion:
        """Return the corresponding Quaternion representation.

        Returns
        -------
        Quaternion
            The equivalent Quaternion representation of this rotation.
        """
        quat = euler_to_quaternion(self.array, seq=self.seq)
        return Quaternion(quat)

    @classmethod
    def from_euler(cls, euler: EulerAngles, seq: str = "xyz") -> EulerAngles:
        """Return an EulerAngles instance from another EulerAngles instance.

        Can be used to change the sequence of axes.
        """
        if seq == euler.seq:
            return cls(euler.array, seq=seq)

        return euler.as_euler(seq)

    def as_euler(self, seq: str = "xyz") -> EulerAngles:
        """Return the Euler angles in a different sequence of axes.

        If the requested sequence is the same as the current sequence, returns self.
        """
        if seq == self.seq:
            return self

        return self.as_quat().as_euler(seq)

    @classmethod
    def identity(cls, seq: str = "xyz") -> EulerAngles:
        """Return the identity EulerAngles (zero rotation).

        Parameters
        ----------
        seq : str, optional
            Sequence of axes for Euler angles. Default is 'xyz'.

        Returns
        -------
        EulerAngles
            New EulerAngles instance representing the identity rotation.
        """
        num_angles = len(seq)
        angles = np.zeros(num_angles)
        return cls(angles, seq=seq)


@struct
class Quaternion:
    """Quaternion representation of a rotation in 3 dimensions.

    This class is closely modeled after [scipy.spatial.transform.Rotation](
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.transform.Rotation.html)
    with a few differences:

    - The quaternion is always represented in scalar-first format
      (i.e. [w, x, y, z]) instead of scalar-last ([x, y, z, w]).
    - This class is designed for symbolic computation, so some checks (e.g. for valid
      rotation matrices) are omitted, since these cannot be done symbolically.
    - The class does not support multiple rotations in a single object
    - This implementation supports kinematic calculations

    The following operations on quaternions are supported:

    - Application on vectors (rotations of vectors)
    - Quaternion Composition
    - Quaternion Inversion
    - Kinematic time derivative given angular velocity

    Parameters
    ----------
    quat : array_like, shape (4,)
        Quaternion representing the rotation in scalar-first format (w, x, y, z).

    Attributes
    ----------
    array : np.ndarray, shape (4,)
        Underlying numpy array representing the quaternion.

    Examples
    --------
    >>> from archimedes.spatial import Quaternion
    >>> import numpy as np

    Consider a counter-clockwise rotation of 90 degrees about the z-axis. This
    corresponds to the following quaternion (in scalar-first format):

    >>> q = Quaternion([np.cos(np.pi/4), 0, 0, np.sin(np.pi/4)])

    The quaternion can be expressed in any of the other formats:

    >>> q.as_matrix()
    array([[ 2.22044605e-16,  1.00000000e+00,  0.00000000e+00],
       [-1.00000000e+00,  2.22044605e-16,  0.00000000e+00],
       [ 0.00000000e+00,  0.00000000e+00,  1.00000000e+00]])
    >>> np.rad2deg(q.as_euler('zyx'))
    array([90.,  0.,  0.])

    The same quaternion can be initialized using a rotation matrix:

    >>> q = Quaternion.from_matrix([[0, 1, 0],
    ...                    [-1, 0, 0],
    ...                    [0, 0, 1]])

    Representation in other formats:

    >>> np.rad2deg(q.as_euler('zyx'))
    array([90.,  0.,  0.])

    The ``from_euler`` method is flexible in the range of input formats
    it supports. Here we initialize a quaternion about a single axis:

    >>> q = Quaternion.from_euler(np.deg2rad(90), 'z')

    The associated rotation matrix can be used to change coordinate systems. If the
    quaternion represents the orientation of a frame B relative to a frame A, then the
    rotation matrix transforms vectors from frame A to frame B:

    >>> v_A = np.array([1, 0, 0])  # Vector in frame A
    >>> R_BA = q.as_matrix()
    >>> R_BA @ v_A  # Vector in frame B
    [6.12323e-17, -1, 0]

    The ``kinematics`` method can be used to compute the time derivative of the
    quaternion as an attitude representation given the angular velocity in the
    body frame using quaternion kinematics:

    >>> w_B = np.array([0, 0, np.pi/2])  # 90 deg/s about z-axis
    >>> q.kinematics(w_B)
    array([-0.55536037,  0.        ,  0.        ,  0.55536037])

    See Also
    --------
    scipy.spatial.transform.Rotation : Similar class in SciPy
    RigidBody : Rigid body dynamics supporting ``Quaternion`` attitude representation
    euler_to_dcm : Directly calculate rotation matrix from roll-pitch-yaw angles
    euler_kinematics : Transform roll-pitch-yaw rates to body-frame angular velocity
    quaternion_kinematics : Low-level quaternion kinematics function
    """

    array: np.ndarray

    def __post_init__(self):
        quat = self.array
        # Internal implementation: for some analysis (e.g. vmap), the
        # struct might be initialized with an int instead of an array
        # for determining axis indices.  In that case we don't want to
        # do explicit shape validation
        # Also for vmapping: data might be passed as a 2D array
        # This solution isn't ideal, since it could introduce user errors
        # by inadvertently passing a 2D array.  It might be better to
        # check the angles when methods are called instead.
        if not (isinstance(quat, int) or getattr(quat, "ndim", None) == 2):
            quat = np.hstack(quat)  # type: ignore
            if quat.shape not in [(4,), (1, 4), (4, 1)]:
                raise ValueError("Quaternion must have shape (4,), (1, 4), or (4, 1)")
            quat = quat.flatten()

        # NOTE: object.__setattr__ is used instead of directly setting because
        # this is a frozen dataclass
        object.__setattr__(self, "array", quat)

    # === Methods for implementing Attitude protocol ===

    def as_matrix(self) -> np.ndarray:
        """Return the quaternion as a rotation matrix.

        If the attitude represents the orientation of a body A relative to a frame B,
        then this method returns the matrix R_BA that transforms vectors from
        frame A to frame B.  Specifically, for a vector v_A expressed in frame A,
        the corresponding vector in frame B is given by ``v_B = R_BA @ v_A``.

        The inverse transformation can be obtained by transposing this matrix:
        ``R_AB = R_BA.T``.

        Returns
        -------
            A 3x3 numpy array representing the DCM.
        """
        return quaternion_to_dcm(self.array)

    def inv(self) -> Quaternion:
        """Return the inverse of the quaternion

        Returns
        -------
        Quaternion
            A new Quaternion instance representing the inverse rotation.
        """
        q_inv = quaternion_inverse(self.array)
        return Quaternion(q_inv)

    def kinematics(self, w: np.ndarray, baumgarte: float | None = None) -> Quaternion:
        """Return the time derivative of the quaternion given angular velocity w.

        If the quaternion represents the attitude of a body B, then w_B should be
        the body relative angular velocity ω_B.

        The derivative is computed using quaternion kinematics:
            dq/dt = 0.5 * q ⊗ [0, ω]
        where ⊗ is the quaternion multiplication operator.

        The method optionally support Baumgarte stabilization to preserve
        unit normalization.  For a stabilization factor λ, the full
        time derivative is:
            dq/dt = 0.5 * q ⊗ [0, ω] - λ * (||q||² - 1) * q

        **CAUTION**: This method returns the time derivative of the attitude,
        which is represented with the same data structure for consistency with
        ODE solving - but this return is not itself a valid rotation representation
        until integrated in time. Hence, the output of ``kinematics`` should never
        be converted to a different attitude representation or rotation matrix.

        Parameters
        ----------
        w : array_like, shape (3,)
            Angular velocity vector in the body frame.
        baumgarte : float, optional
            Baumgarte stabilization factor. If > 0, Baumgarte stabilization is
            applied to enforce unit norm constraint. Default is 0 (no stabilization).

        Returns
        -------
        Quaternion
            The time derivative represented as a Quaternion instance.
        """
        q_dot = quaternion_kinematics(self.array, w, baumgarte=baumgarte)
        return Quaternion(q_dot)

    # === Other methods ===

    def __repr__(self):
        return f"Quaternion({self.array})"

    def __len__(self):
        return len(self.array)

    def __getitem__(self, index: int):
        return self.array[index]

    def __iter__(self):
        return iter(self.array)

    @classmethod
    def from_quat(cls, quat: Quaternion) -> Quaternion:
        """Returns a copy of the Quaternion object - dummy method for API consistency.

        Returns
        -------
        Quaternion
            A copy of the input Quaternion instance.
        """
        return cls(quat.array)

    def as_quat(self) -> Quaternion:
        """Return the same object - dummy method for API consistency.

        Returns
        -------
        Quaternion
            The same Quaternion instance.
        """
        return self

    @classmethod
    def from_matrix(cls, matrix: np.ndarray) -> Quaternion:
        """Create a Quaternion from a rotation matrix.

        Note that for the sake of symbolic computation, this method assumes that
        the input is a valid rotation matrix (orthogonal and determinant +1).

        Parameters
        ----------
        matrix : array_like, shape (3, 3)
            Quaternion matrix.

        Returns
        -------
        Quaternion
            A new Quaternion instance.

        See Also
        --------
        dcm_to_quaternion : Low-level direction cosine matrix to quaternion conversion
        """
        quat = dcm_to_quaternion(matrix)
        return cls(quat)

    @classmethod
    def from_euler(
        cls, euler: EulerAngles | np.ndarray, seq: str | None = None
    ) -> Quaternion:
        """Create a Quaternion from Euler angles.

        Parameters
        ----------
        euler : EulerAngles or array_like
            Euler angles instance or array of Euler angles in radians.
        seq : str, optional
            Sequence of axes for Euler angles (up to length 3).  Each character must
            be one of 'x', 'y', 'z' (extrinsic) or 'X', 'Y', 'Z' (intrinsic).  Default
            is 'xyz'. Should not be specified if `euler` is an EulerAngles instance.

        Returns
        -------
        Quaternion
            A new Quaternion instance.

        See Also
        --------
        euler_to_quaternion : Low-level Euler to quaternion conversion function
        """

        if isinstance(euler, EulerAngles):
            if seq is not None:
                raise ValueError(
                    "If `euler` is an EulerAngles instance, `seq` should not be passed"
                )
        else:
            if seq is None:
                seq = "xyz"
            euler = EulerAngles(euler, seq=seq)

        return euler.as_quat()

    def as_euler(self, seq: str) -> EulerAngles:
        """Return the Euler angles from the quaternion

        This method uses the same notation and conventions as the SciPy Rotation class.
        See the SciPy documentation and :py:meth:``from_euler`` for more details.

        See Also
        --------
        quaternion_to_euler : Low-level quaternion to Euler conversion function
        """
        return EulerAngles.from_quat(self, seq=seq)

    @classmethod
    def identity(cls) -> Quaternion:
        """Return a quaternion representing the identity rotation."""
        return cls(np.array([1.0, 0.0, 0.0, 0.0]))

    def mul(self, other: Quaternion, normalize: bool = False) -> Quaternion:
        """Compose (multiply) this quaternion with another"""
        q1 = self.array
        q2 = other.array
        q = quaternion_multiply(q1, q2)
        if normalize:
            q = q / np.linalg.norm(q)
        return Quaternion(q)

    def __mul__(self, other: Quaternion) -> Quaternion:
        """Compose (multiply) this quaternion with another and normalize the result"""
        return self.mul(other, normalize=True)



# ruff: noqa: N803, N806, N816
from __future__ import annotations

import abc

import numpy as np

import archimedes as arc
from archimedes import StructConfig, UnionConfig, struct

__all__ = [
    "F16Engine",
    "F16EngineConfig",
    "TabulatedEngine",
    "TabulatedEngineConfig",
    "NASAEngine",
    "NASAEngineConfig",
]

#
# Engine lookup tables
#
alt_vector = np.array([0.0, 10000.0, 20000.0, 30000.0, 40000.0, 50000.0])
mach_vector = np.array([0.0, 0.2, 0.4, 0.6, 0.8, 1.0])

Tidl_data = np.array(
    [
        1060.0,
        670.0,
        880.0,
        1140.0,
        1500.0,
        1860.0,
        635.0,
        425.0,
        690.0,
        1010.0,
        1330.0,
        1700.0,
        60.0,
        25.0,
        345.0,
        755.0,
        1130.0,
        1525.0,
        -1020.0,
        -710.0,
        -300.0,
        350.0,
        910.0,
        1360.0,
        -2700.0,
        -1900.0,
        -1300.0,
        -247.0,
        600.0,
        1100.0,
        -3600.0,
        -1400.0,
        -595.0,
        -342.0,
        -200.0,
        700.0,
    ]
).reshape((6, 6), order="F")

Tmil_data = np.array(
    [
        12680.0,
        9150.0,
        6200.0,
        3950.0,
        2450.0,
        1400.0,
        12680.0,
        9150.0,
        6313.0,
        4040.0,
        2470.0,
        1400.0,
        12610.0,
        9312.0,
        6610.0,
        4290.0,
        2600.0,
        1560.0,
        12640.0,
        9839.0,
        7090.0,
        4660.0,
        2840.0,
        1660.0,
        12390.0,
        10176.0,
        7750.0,
        5320.0,
        3250.0,
        1930.0,
        11680.0,
        9848.0,
        8050.0,
        6100.0,
        3800.0,
        2310.0,
    ]
).reshape((6, 6), order="F")

Tmax_data = np.array(
    [
        20000.0,
        15000.0,
        10800.0,
        7000.0,
        4000.0,
        2500.0,
        21420.0,
        15700.0,
        11225.0,
        7323.0,
        4435.0,
        2600.0,
        22700.0,
        16860.0,
        12250.0,
        8154.0,
        5000.0,
        2835.0,
        24240.0,
        18910.0,
        13760.0,
        9285.0,
        5700.0,
        3215.0,
        26070.0,
        21075.0,
        15975.0,
        11115.0,
        6860.0,
        3950.0,
        28886.0,
        23319.0,
        18300.0,
        13484.0,
        8642.0,
        5057.0,
    ]
).reshape((6, 6), order="F")

Tidl_interpolant = arc.interpolant([alt_vector, mach_vector], Tidl_data)
Tmil_interpolant = arc.interpolant([alt_vector, mach_vector], Tmil_data)
Tmax_interpolant = arc.interpolant([alt_vector, mach_vector], Tmax_data)


@struct
class F16Engine(metaclass=abc.ABCMeta):
    lo_gear: float = 64.94  # Low gear throttle slope
    hi_gear: float = 217.38  # High gear throttle slope
    throttle_breakpoint: float = 0.77  # Switch between linear throttle models

    @struct
    class Input:
        throttle: float  # Throttle position [0-1]
        alt: float  # Altitude [ft]
        mach: float  # Mach number

    @struct
    class Output:
        thrust: np.ndarray  # Thrust magnitude [lbf]

    @struct
    class State:
        pass  # No state by default

    def tgear(self, thtl):
        c_hi = (self.hi_gear - self.lo_gear) * self.throttle_breakpoint
        return np.where(
            thtl <= self.throttle_breakpoint,
            self.lo_gear * thtl,
            self.hi_gear * thtl - c_hi,
        )

    def dynamics(self, t: float, x: State, u: Input) -> State:
        """Time derivative of engine model state"""
        return x  # No dynamics by default

    @abc.abstractmethod
    def output(self, t: float, x: State, u: Input) -> Output:
        """Calculate engine thrust output"""
        pass

    @abc.abstractmethod
    def trim(self, throttle: float) -> State:
        """Calculate trim conditions for the engine model"""
        pass


class EngineConfigBase(StructConfig):
    lo_gear: float = 64.94  # Low gear throttle slope
    hi_gear: float = 217.38  # High gear throttle slope
    throttle_breakpoint: float = 0.77  # Switch between linear throttle models


@struct
class TabulatedEngine(F16Engine):
    """Simple tabulated engine model with zero dynamics"""

    def _calc_thrust(self, power: float, alt: float, mach: float) -> np.ndarray:
        """Calculate body-frame thrust vector from engine power"""
        T_mil = Tmil_interpolant(alt, mach)
        T_idl = Tidl_interpolant(alt, mach)
        T_max = Tmax_interpolant(alt, mach)

        return np.where(
            power < 50.0,
            T_idl + (T_mil - T_idl) * power * 0.02,
            T_mil + (T_max - T_mil) * (power - 50.0) * 0.02,
        )

    def output(
        self, t: float, x: F16Engine.State, u: F16Engine.Input
    ) -> F16Engine.Output:
        power = self.tgear(u.throttle)
        thrust = self._calc_thrust(power, u.alt, u.mach)
        return F16Engine.Output(thrust=thrust)

    def trim(self, throttle: float) -> F16Engine.State:
        return F16Engine.State()  # No state to trim


class TabulatedEngineConfig(EngineConfigBase, type="tabulated"):
    def build(self) -> TabulatedEngine:
        return TabulatedEngine(
            lo_gear=self.lo_gear,
            hi_gear=self.hi_gear,
            throttle_breakpoint=self.throttle_breakpoint,
        )


@struct
class NASAEngine(TabulatedEngine):
    """Tabulated engine model with first-order lag dynamics based on NASA model"""

    rtau_min: float = 0.1  # Minimum inv time constant for engine response [1/s]
    rtau_max: float = 1.0  # Maximum inv time constant for engine response [1/s]

    @struct
    class State(F16Engine.State):
        power: float  # Engine power

    def _rtau(self, dP):
        """Inverse time constant for engine response"""
        return np.where(
            dP <= 25,
            self.rtau_max,
            np.where(dP >= 50, self.rtau_min, 1.9 - 0.036 * dP),
        )

    def dynamics(
        self, t: float, x: NASAEngine.State, u: F16Engine.Input
    ) -> NASAEngine.State:
        """Time derivative of engine model state"""
        P = x.power  # Engine power
        thtl = u.throttle  # Throttle position

        cpow = self.tgear(thtl)  # Command power
        P2 = np.where(
            cpow >= 50.0,
            np.where(P >= 50.0, cpow, 60.0),
            np.where(P >= 50.0, 40.0, cpow),
        )

        # 1/tau
        rtau = np.where(P >= 50.0, 5.0, self._rtau(P2 - P))

        dP_dt = rtau * (P2 - P)
        return NASAEngine.State(power=dP_dt)

    def output(
        self, t: float, x: NASAEngine.State, u: F16Engine.Input
    ) -> F16Engine.Output:
        thrust = self._calc_thrust(x.power, u.alt, u.mach)
        return F16Engine.Output(thrust=thrust)

    def trim(self, throttle: float) -> NASAEngine.State:
        power = self.tgear(throttle)
        return NASAEngine.State(power=power)


class NASAEngineConfig(EngineConfigBase, type="nasa"):
    rtau_min: float = 0.1  # Minimum inv time constant for engine response [1/s]
    rtau_max: float = 1.0  # Maximum inv time constant for engine response [1/s]

    def build(self) -> NASAEngine:
        return NASAEngine(
            lo_gear=self.lo_gear,
            hi_gear=self.hi_gear,
            throttle_breakpoint=self.throttle_breakpoint,
            rtau_min=self.rtau_min,
            rtau_max=self.rtau_max,
        )


F16EngineConfig = UnionConfig[
    TabulatedEngineConfig,
    NASAEngineConfig,
]



"""Utilities for working with hierarchical tree-structured data."""

from ._config import (
    StructConfig,
    UnionConfig,
)
from ._flatten_util import ravel_tree as ravel
from ._registry import (
    register_dataclass,
    register_struct,
)
from ._struct import (
    InitVar,
    field,
    fields,
    is_struct,
    replace,
    struct,
)
from ._tree_util import is_leaf
from ._tree_util import (
    tree_all as all,
)
from ._tree_util import (
    tree_flatten as flatten,
)
from ._tree_util import (
    tree_leaves as leaves,
)
from ._tree_util import (
    tree_map as map,
)
from ._tree_util import (
    tree_reduce as reduce,
)
from ._tree_util import (
    tree_structure as structure,
)
from ._tree_util import (
    tree_unflatten as unflatten,
)

__all__ = [
    "register_struct",
    "register_dataclass",
    "is_leaf",
    "flatten",
    "unflatten",
    "structure",
    "leaves",
    "map",
    "all",
    "reduce",
    "ravel",
    "struct",
    "field",
    "InitVar",
    "is_struct",
    "fields",
    "replace",
    "StructConfig",
    "UnionConfig",
]



"""Low-level functions for Euler angles.

These functions are for conversions and kinematics and operate directly on
arrays rather than higher-level wrapper classes.
"""

# ruff: noqa: N806, N803, N815
from __future__ import annotations

import re
from typing import cast

import numpy as np

from archimedes import array

__all__ = [
    "euler_kinematics",
    "euler_to_dcm",
]


def _check_seq(seq: str) -> bool:
    num_axes = len(seq)
    if num_axes < 1 or num_axes > 3:
        raise ValueError(
            "Expected axis specification to be a non-empty "
            "string of upto 3 characters, got {}".format(seq)
        )

    # The following checks are verbatim from:
    # https://github.com/scipy/scipy/blob/3ead2b543df7c7c78619e20f0cb6139e344a8866/scipy/spatial/transform/_rotation_cy.pyx#L461-L476  # ruff: noqa: E501
    intrinsic = re.match(r"^[XYZ]{1,3}$", seq) is not None
    extrinsic = re.match(r"^[xyz]{1,3}$", seq) is not None
    if not (intrinsic or extrinsic):
        raise ValueError(
            "Expected axes from `seq` to be from ['x', 'y', "
            "'z'] or ['X', 'Y', 'Z'], got {}".format(seq)
        )

    if any(seq[i] == seq[i + 1] for i in range(len(seq) - 1)):
        raise ValueError(
            "Expected consecutive axes to be different, got {}".format(seq)
        )

    return intrinsic


def _check_angles(angles: np.ndarray, seq: str) -> np.ndarray:
    num_axes = len(seq)

    if isinstance(angles, (list, tuple)):
        angles = np.hstack(angles)

    angles = np.atleast_1d(angles)
    if angles.shape not in [(num_axes,), (1, num_axes), (num_axes, 1)]:
        raise ValueError(
            f"For {seq} sequence with {num_axes} axes, `angles` must have shape "
            f"({num_axes},), (1, {num_axes}), or ({num_axes}, 1). Got "
            f"{angles.shape}"
        )
    return angles


def _rot_x(angle: float) -> np.ndarray:
    """Rotation about x-axis by given angle (radians)."""
    c = np.cos(angle)
    s = np.sin(angle)

    R = array(
        [
            [1.0, 0.0, 0.0],
            [0.0, c, s],
            [0.0, -s, c],
        ]
    )
    return cast(np.ndarray, R)


def _rot_y(angle: float) -> np.ndarray:
    """Rotation about y-axis by given angle (radians)."""
    c = np.cos(angle)
    s = np.sin(angle)

    R = array(
        [
            [c, 0.0, -s],
            [0.0, 1.0, 0.0],
            [s, 0.0, c],
        ]
    )
    return cast(np.ndarray, R)


def _rot_z(angle: float) -> np.ndarray:
    """Rotation about z-axis by given angle (radians)."""
    c = np.cos(angle)
    s = np.sin(angle)

    R = array(
        [
            [c, s, 0.0],
            [-s, c, 0.0],
            [0.0, 0.0, 1.0],
        ]
    )
    return cast(np.ndarray, R)


def euler_to_dcm(angles: np.ndarray, seq: str = "xyz") -> np.ndarray:
    """Direction cosine matrix from Euler angles

    If the Euler angles represent the attitude of a body B relative to a frame A,
    then this function returns the matrix R_BA that transforms vectors from
    frame A to frame B.  Specifically, for a vector v_A expressed in frame A,
    the corresponding vector in frame B is given by ``v_B = R_BA @ v_A``.

    The inverse transformation can be obtained by transposing this matrix:
    ``R_AB = R_BA.T``.

    By default, the Euler angle sequence is assumed to follow the standard aerospace
    convention of an extrinsic roll-pitch-yaw sequence ("xyz").  However, it supports
    arbitrary sequences of non-repeating axes up to length 3. Both intrinsic
    (uppercase letters) and extrinsic (lowercase letters) sequences are supported.

    In general, the ``Quaternion`` class should be preferred over Euler representations,
    although Euler angles are used in some special cases (e.g. stability analysis).
    In these cases, this function gives a more direct calculation of the
    transformation matrix without converting to the intermediate quaternion.

    Parameters
    ----------
    angles : array_like
        Euler angles in radians representing the orientation of frame B with respect to
        frame A. Shape must match the length of ``seq``.
    seq : str, optional
        Sequence of axes for Euler angles (up to length 3).  Each character must be one
        of 'x', 'y', 'z' (extrinsic) or 'X', 'Y', 'Z' (intrinsic).  Default is 'xyz'.

    Returns
    -------
    np.ndarray, shape (3, 3)
        Direction cosine matrix R_BA that transforms vectors from frame A to frame B.
    """

    # Validate angle sequence
    intrinsic = _check_seq(seq)
    angles = _check_angles(angles, seq)

    seq = seq.lower()
    if intrinsic:
        seq = seq[::-1]  # Reverse for intrinsic rotations
        angles = angles[::-1]

    # Note that this approach of building the DCM by composing
    # elemental rotations is usually slower than the direct formula,
    # since it involves multiple matrix multiplications. However, with
    # symbolic arrays there is no difference in speed because the
    # multiplications are not actually carried out until after the
    # full matrix is built.
    R = np.eye(3, like=angles)
    for char, angle in zip(seq, angles):
        match char:
            case "x":
                R = R @ _rot_x(angle)
            case "y":
                R = R @ _rot_y(angle)
            case "z":
                R = R @ _rot_z(angle)
    return R


def euler_kinematics(rpy: np.ndarray, inverse: bool = False) -> np.ndarray:
    """Euler kinematical equations

    Defining 𝚽 = [phi, theta, psi] == Euler angles for roll, pitch, yaw
    attitude representation, this function returns a matrix H(𝚽) such
    that
        d𝚽/dt = H(𝚽) * ω.

    If inverse=True, it returns a matrix H(𝚽)^-1 such that
        ω = H(𝚽)^-1 * d𝚽/dt.

    This function supports _only_ the extrinsic roll-pitch-yaw sequence ("xyz").

    Parameters
    ----------
    rpy : array_like, shape (3,)
        Roll, pitch, yaw angles in radians.
    inverse : bool, optional
        If True, returns the inverse matrix H(𝚽)^-1. Default is False.

    Returns
    -------
    np.ndarray, shape (3, 3)
        The transformation matrix H(𝚽) or its inverse.

    Notes
    -----

    Typical rigid body dynamics calculations provide the body-frame angular velocity
    ω_B, but this is _not_ the time derivative of the Euler angles.  Instead, one
    can define a matrix H(𝚽) such that d𝚽/dt = H(𝚽) * ω_B.

    This matrix H(𝚽) has a singularity at θ = ±π/2 (gimbal lock).

    Note that the ``RigidBody`` class by default uses quaternions (via the
    ``Quaternion`` class) for attitude representation.
    In general this is preferred due to the gimbal lock singularity, but
    special cases like stability analysis may use Euler angle kinematics.
    """

    φ, θ = rpy[0], rpy[1]  # Roll, pitch

    sφ, cφ = np.sin(φ), np.cos(φ)
    sθ, cθ = np.sin(θ), np.cos(θ)
    tθ = np.tan(θ)

    _1 = np.ones_like(φ)
    _0 = np.zeros_like(φ)

    if inverse:
        Hinv = np.array(
            [
                [_1, _0, -sθ],
                [_0, cφ, cθ * sφ],
                [_0, -sφ, cθ * cφ],
            ],
            like=rpy,
        )
        return Hinv

    else:
        H = np.array(
            [
                [_1, sφ * tθ, cφ * tθ],
                [_0, cφ, -sφ],
                [_0, sφ / cθ, cφ / cθ],
            ],
            like=rpy,
        )
        return H



import abc
import dataclasses
from typing import Callable

import numpy as np

from archimedes import compile, minimize, struct, field

from .discretization import SplineDiscretization
from .interpolation import LagrangePolynomial


@dataclasses.dataclass
class BoundaryData:
    t0: float
    x0: np.ndarray
    tf: float
    xf: np.ndarray
    p: np.ndarray


@dataclasses.dataclass
class Constraint:
    func: Callable
    nc: int
    lower_bound: np.ndarray = None  # Default to equality constraint
    upper_bound: np.ndarray = None

    def __post_init__(self):
        if self.lower_bound is None:
            self.lower_bound = np.zeros(self.nc)
        if self.upper_bound is None:
            self.upper_bound = np.zeros(self.nc)

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)


#
# Convenience functions for boundary conditions
#
def initial_condition(x0):
    return Constraint(lambda data: data.x0 - x0, len(x0))


def final_condition(xf):
    return Constraint(lambda data: data.xf - xf, len(xf))


def start_time(t0):
    return Constraint(lambda data: data.t0 - t0, 1)


def end_time(tf):
    return Constraint(lambda data: data.tf - tf, 1)


def parameter_bounds(lb, ub):
    if len(lb) != len(ub):
        raise ValueError("Lower and upper bounds must have the same length")
    return Constraint(lambda data: data.p, len(lb), lower_bound=lb, upper_bound=ub)


#
# Convenience functions for constructing path constraints
#
def state_bounds(lb, ub):
    if len(lb) != len(ub):
        raise ValueError("Lower and upper bounds must have the same length")
    return Constraint(lambda t, x, u, p: x, len(lb), lower_bound=lb, upper_bound=ub)


def control_bounds(lb, ub):
    if len(lb) != len(ub):
        raise ValueError("Lower and upper bounds must have the same length")
    return Constraint(lambda t, x, u, p: u, len(lb), lower_bound=lb, upper_bound=ub)


@struct
class OptimalControlSolution:
    xp: np.ndarray
    up: np.ndarray
    tp: np.ndarray
    x: Callable = field(static=True)
    u: Callable = field(static=True)
    p: np.ndarray
    dvs: np.ndarray = None

    @property
    def t0(self):
        return self.tp[0]

    @property
    def tf(self):
        return self.tp[-1]


class OCPBase(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def build_objective(self, domain):
        """Construct the objective function"""

    @abc.abstractmethod
    def build_constraints(self, domain):
        """Construct the constraint functions"""

    @abc.abstractmethod
    def initialize(self, domain, t0, tf, x_guess=None, u_guess=None):
        """Initialize the decision variables and bounds"""

    @abc.abstractmethod
    def postprocess(self, sol, domain):
        """Postprocess the solution"""

    @abc.abstractmethod
    def dynamics_residual(self, sol, element, x, t0, tf):
        """Compute the residual of the dynamics for a given solution"""

    def solve(self, domain, t_guess=None, x_guess=None, u_guess=None, options=None):
        initial_guess, lower, upper = self.initialize(domain, t_guess, x_guess, u_guess)
        obj = self.build_objective(domain)
        cons = self.build_constraints(domain)
        opt_result = minimize(
            obj,
            initial_guess,
            constr=cons,
            constr_bounds=(lower, upper),
            options=options,
        )
        opt_dvs = opt_result.x
        return self.postprocess(opt_dvs, domain)


@dataclasses.dataclass
class OptimalControlProblem(OCPBase):
    nx: int
    nu: int
    ode: Callable
    quad: Callable
    cost: Callable
    np: int = 0
    boundary_constraints: list[Constraint] = dataclasses.field(default_factory=list)
    path_constraints: list[Constraint] = dataclasses.field(default_factory=list)

    def num_dvs(self, domain):
        N = domain.n_nodes
        nx, nu = self.nx, self.nu
        return (N + 1) * nx + N * nu + 2 + self.np

    def unpack_dvs(self, dvs, domain, order="C"):
        # Note on ordering: CasADi orders the symbolic variables in Fortran style,
        # but the NumPy default is C-style.  So the default behavior is to unpack
        # with F ordering in the objective and constraint functions, but then use
        # C ordering by default in the unpacking function for postprocessing, etc.
        N = domain.n_nodes
        nx, nu = self.nx, self.nu
        i0, i1 = 0, (N + 1) * nx
        x = np.reshape(dvs[i0:i1], (N + 1, nx), order=order)
        i0, i1 = i1, i1 + N * nu
        u = np.reshape(dvs[i0:i1], (N, nu), order=order)
        i0, i1 = i1, i1 + self.np
        p = dvs[i0:i1]
        t0, tf = dvs[-2:]  # Time knots
        return x, u, t0, tf, p

    def boundary_data(self, dvs, domain):
        x, u, t0, tf, p = self.unpack_dvs(dvs, domain, order="C")
        return BoundaryData(t0, x[0, :].T, tf, x[-1, :].T, p)

    def build_objective(self, domain):
        N, w = domain.n_nodes, domain.weights

        # Given a flattened array of decision variables, compute the
        # value of the objective function
        @compile
        def obj(dvs):
            x, u, t0, tf, p = self.unpack_dvs(dvs, domain, order="C")

            tscale = 0.5 * (tf - t0)
            t = domain.time_nodes(t0, tf)

            running_cost = []
            for i in range(N):
                running_cost = np.append(
                    running_cost, self.quad(t[i], x[i, :], u[i, :], p)
                )

            # Evaluate objective
            q = tscale * np.dot(w, running_cost)
            J = self.cost(x[0, :], t0, x[-1, :], tf, q, p)
            return J

        return obj

    def build_constraints(self, domain):
        N, D = domain.n_nodes, domain.diff_matrix

        # Given a flattened array of decision variables, compute the
        # function value of the constraints
        @compile
        def cons(dvs):
            x, u, t0, tf, p = self.unpack_dvs(dvs, domain, order="C")

            t = domain.time_nodes(t0, tf)
            tscale = 0.5 * (tf - t0)

            x_dot = (
                D @ x
            )  # Numerically differentiate with Lagrange differentiation matrix

            # Dynamical constraints
            # TODO: Support mass matrix here for DAEs
            res = []
            # F = tscale * self.ode(t[:-1], x[:-1, :].T, u.T, p)

            # vmap_ode = vmap(self.ode, in_axes=(0, 1, 1, None), out_axes=1)
            # print(t[:-1].shape, x[:-1, :].T.shape, u.T.shape)
            # F = tscale * vmap_ode(t[:-1], x[:-1, :].T, u.T, p)
            for i in range(N):
                # res = np.append(res, x_dot[i, :].T - F[:, i])
                F = tscale * self.ode(t[i], x[i, :], u[i, :], p)
                res = np.append(res, x_dot[i, :].T - F)

            # # Constrain the non-collocated end point
            # w = domain.weights
            # res = np.append(res, x[-1] - (x[0] + F @ w))

            # Boundary condition constraints
            boundary_data = self.boundary_data(dvs, domain)
            for bc in self.boundary_constraints:
                res = np.append(res, bc(boundary_data))

            # Path constraints
            for pc in self.path_constraints:
                for i in range(N):
                    res = np.append(res, pc(t[i], x[i, :], u[i, :], p))

            res = np.append(res, tf - t0)  # Ordering of endpoints

            return res

        return cons

    def initialize(
        self, domain, t_guess=None, x_guess=None, u_guess=None, p_guess=None
    ):
        N = domain.n_nodes
        nx, nu = self.nx, self.nu

        if t_guess is None:
            t_guess = [0.0, 1.0]

        if p_guess is None:
            p_guess = np.zeros(self.np)

        t0, tf = t_guess
        t = domain.time_nodes(t0, tf)

        if x_guess is None:
            x_guess = lambda t: np.zeros(nx)

        if u_guess is None:
            u_guess = lambda t: np.zeros(nu)

        x0 = np.vstack([x_guess(t[i]) for i in range(N + 1)])
        u0 = np.vstack([u_guess(t[i]) for i in range(N)])

        initial_guess = np.hstack(
            [x0.flatten(), u0.flatten(), p_guess.flatten(), t0, tf]
        )

        # Initialize constraint vectors for dynamic constraints
        # lb = np.zeros(nx * (N + 1))
        # ub = np.zeros(nx * (N + 1))
        lb = np.zeros(nx * N)
        ub = np.zeros(nx * N)

        # Boundary condition constraints
        for bc in self.boundary_constraints:
            lb = np.append(lb, bc.lower_bound)
            ub = np.append(ub, bc.upper_bound)

        # Path constraints
        for pc in self.path_constraints:
            lb = np.append(lb, np.tile(pc.lower_bound, N))
            ub = np.append(ub, np.tile(pc.upper_bound, N))

        # Time ordering constraint
        lb = np.append(lb, 0.0)
        ub = np.append(ub, np.inf)

        return initial_guess, lb, ub

    def postprocess(self, sol, domain: SplineDiscretization):
        x, u, t0, tf, p = self.unpack_dvs(sol, domain, order="C")
        t = domain.time_nodes(t0, tf)
        x_fn, u_fn = domain.create_interpolants(x, u, t0, tf)
        return OptimalControlSolution(x, u, t, x_fn, u_fn, p, dvs=sol)

    def dynamics_residual(self, sol, element, x, t0, tf):
        # Differentiate then interpolate
        D = element.diff_matrix
        x_dot_ = D @ x

        # Extrapolate to end time
        τ = element.nodes
        p_radau = LagrangePolynomial(τ[:-1])  # Interpolant for collocated nodes
        x_dot_f = p_radau.interpolate(x_dot_, τ[-1])
        x_dot_ = np.vstack([x_dot_, x_dot_f])
        x_dot_fn = element.create_interpolant(x_dot_, t0, tf)

        def _res(t):
            x = sol.x(t)
            u = sol.u(t)
            F = np.zeros_like(x)
            for i in range(len(x)):
                F[i] = self.ode(t[i], x[i], u[i], sol.p)
            x_dot = x_dot_fn(t)
            return abs(x_dot - 0.5 * (tf - t0) * F)

        return _res


@dataclasses.dataclass
class MultiStageOptimalControlProblem(OCPBase):
    stages: list[OptimalControlProblem]
    stage_constraints: list[Constraint]

    def split_dvs(self, dvs, domain):
        num_dvs = [s.num_dvs(d) for (s, d) in zip(self.stages, domain)]

        # TODO: Implement np.split for symbolic arrays
        # return np.split(dvs, np.cumsum(num_dvs)[:-1])

        split_dvs = []
        idx = 0
        for n in num_dvs:
            split_dvs.append(dvs[idx : idx + n])
            idx += n
        return split_dvs

    def build_objective(self, domain: list[SplineDiscretization]):
        sub_objectives = [s.build_objective(d) for (s, d) in zip(self.stages, domain)]

        @compile
        def obj(dvs):
            # First split dvs by stage
            split_dvs = self.split_dvs(dvs, domain)

            # Sum the objectives of each stage
            return sum(_obj(x) for (_obj, x) in zip(sub_objectives, split_dvs))

        return obj

    def build_constraints(self, domains: list[SplineDiscretization]):
        sub_constraints = [
            s.build_constraints(d) for (s, d) in zip(self.stages, domains)
        ]

        @compile
        def cons(dvs):
            # First split dvs by stage
            split_dvs = self.split_dvs(dvs, domains)

            # Concatenate the constraints of each stage
            res = []
            for _cons, x in zip(sub_constraints, split_dvs):
                res = np.append(res, _cons(x))

            # Then add the multi-stage constraints
            boundary_data = []
            for s, d, x in zip(self.stages, domains, split_dvs):
                boundary_data.append(s.boundary_data(x, d))

            for sc in self.stage_constraints:
                res = np.append(res, sc(boundary_data))

            # Join the end time of each stage to the start time of the next stage
            for i in range(len(self.stages) - 1):
                tf_prev = boundary_data[i].tf
                t0_next = boundary_data[i + 1].t0
                res = np.append(res, tf_prev - t0_next)

            return res

        return cons

    def initialize(
        self,
        domain: list[SplineDiscretization],
        t_guess: list[float] = None,
        x_guess=None,
        u_guess=None,
    ):
        initial_guess = []
        lower_bound = []
        upper_bound = []

        if t_guess is None:
            t_guess = np.linspace(0, 1, len(self.stages) + 1)

        # Initial guess, lower bound, and upper bound for each stage
        for i, (s, d) in enumerate(zip(self.stages, domain)):
            t0, tf = t_guess[i], t_guess[i + 1]
            (ig, lb, ub) = s.initialize(
                d, t_guess=(t0, tf), x_guess=x_guess, u_guess=u_guess
            )
            initial_guess = np.append(initial_guess, ig)
            lower_bound = np.append(lower_bound, lb)
            upper_bound = np.append(upper_bound, ub)

        # Add stage constraints
        for sc in self.stage_constraints:
            lower_bound = np.append(lower_bound, sc.lower_bound)
            upper_bound = np.append(upper_bound, sc.upper_bound)

        # Add time continuity constraints
        for _ in range(len(self.stages) - 1):
            lower_bound = np.append(lower_bound, 0)
            upper_bound = np.append(upper_bound, 0)

        return (initial_guess, lower_bound, upper_bound)

    def postprocess(self, sol, domain):
        split_dvs = self.split_dvs(sol, domain)
        stage_solns = []
        for i in range(len(split_dvs)):
            stage_solns.append(self.stages[i].postprocess(split_dvs[i], domain[i]))

        return stage_solns

    def dynamics_residual(self, sol, element, x, t0, tf):
        raise NotImplementedError("TODO: implement for multi-stage OCPs")



# ruff: noqa: N806
from pathlib import Path

import numpy as np
import numpy.testing as npt
import pytest
import yaml
from f16 import GRAV_FTS2, SubsonicF16
from stability import (
    LateralInput,
    LateralState,
    LongitudinalInput,
    LongitudinalState,
    StabilityInput,
    StabilityState,
)

import archimedes as arc
from archimedes.spatial import Quaternion, euler_kinematics

CURRENT_PATH = Path(__file__).parent


@pytest.fixture
def trim_cases():
    with open(CURRENT_PATH / "_static" / "trim_cases.yaml", "r") as f:
        trim_cases = yaml.safe_load(f)
    return trim_cases


@pytest.fixture
def f16():
    return SubsonicF16(xcg=0.35)


g0 = GRAV_FTS2  # ft/s^2


def test_352(f16: SubsonicF16):
    """Compare to Table 3.5-2 in Lewis, Johnson, Stevens"""
    f16 = f16.replace(xcg=0.40)

    u = f16.Input(
        throttle=0.9,
        elevator=20.0,
        aileron=-15.0,
        rudder=-20.0,
    )

    # Original state used (Vt, alpha, beta) = (500.0, 0.5, -0.2)
    # New model uses equivalent (u, v, w) = (430.0447, -99.3347, 234.9345)
    #   --> (du, dv, dw) = 100.8536, -218.3080, -437.0399
    p_N = np.array([1000.0, 900.0, -10000.0])  # NED-frame position
    rpy = np.array([-1.0, 1.0, -1.0])  # Roll, pitch, yaw
    v_B = np.array([430.0447, -99.3347, 234.9345])  # Velocity in body frame
    w_B = np.array([0.7, -0.8, 0.9])  # Angular velocity in body frame

    att = Quaternion.from_euler(rpy)
    x_eng = f16.engine.State(90.0)  # Engine power
    x = f16.State(p_N, att, v_B, w_B, x_eng)

    # NOTE: There is a typo in the chapter 3 code implementation of the DCM,
    # leading to a sign change for yaw rate xd[11].  Hence, Table 3.5-2 has
    # 248.1241 instead of -248.1241 (the latter is consistent with the SciPy
    # DCM implementation).
    dp_N_ex = np.array(
        [
            342.4439,  # x (north)
            -266.7707,  # y (east)
            -248.1241,  # z (down)
        ]
    )
    dv_B_ex = np.array(
        [
            100.8536,  # u
            -218.3080,  # v
            -437.0399,  # w
        ]
    )
    dw_B_ex = np.array(
        [
            12.62679,  # p
            0.9649671,  # q
            0.5809759,  # r
        ]
    )

    x_t = f16.dynamics(0.0, x, u)

    # Extract body angular velocity from quaternion derivative
    # q_t = 0.5 * q ⊗ ω_B
    # => ω_B = 2 * q⁻¹ ⊗ q_t
    # This gives a check on the quaternion derivative calculation
    # without using the roll-pitch-yaw rates.
    w_B_out = 2 * (att.inv().mul(x_t.att, normalize=False)).array[1:]

    npt.assert_allclose(x_t.pos, dp_N_ex, atol=1e-2)
    npt.assert_allclose(w_B_out, w_B, atol=1e-2)
    npt.assert_allclose(x_t.v_B, dv_B_ex, atol=1e-2, rtol=1e-3)
    npt.assert_allclose(x_t.w_B, dw_B_ex, atol=1e-2, rtol=1e-3)


def test_362(f16: SubsonicF16):
    """Trim conditions (Sec. 3.6-2 in Lewis, Johnson, Stevens)"""

    vt = 5.020000e2
    alpha = 2.392628e-1
    beta = 5.061803e-4
    u = vt * np.cos(alpha) * np.cos(beta)
    v = vt * np.sin(beta)
    w = vt * np.sin(alpha) * np.cos(beta)

    p_N = np.array([0.0, 0.0, 0.0])  # NED-frame position
    rpy = np.array([1.366289e0, 5.000808e-2, 2.340769e-1])  # Roll, pitch, yaw
    v_B = np.array([u, v, w])  # Velocity in body frame
    w_B = np.array(
        [-1.499617e-2, 2.933811e-1, 6.084932e-2]
    )  # Angular velocity in body frame

    att = Quaternion.from_euler(rpy)
    x_eng = f16.engine.State(6.412363e1)  # Engine power
    x = f16.State(p_N, att, v_B, w_B, x_eng)

    u = f16.Input(
        throttle=8.349601e-1,
        elevator=-1.481766e0,
        aileron=9.553108e-2,
        rudder=-4.118124e-1,
    )

    x_t = f16.dynamics(0.0, x, u)
    assert np.allclose(x_t.v_B, 0.0, atol=1e-4)
    assert np.allclose(x_t.w_B, 0.0, atol=1e-4)

    # Check that the angle rates are correct
    # First we have to convert the desired angular rates to angular momentum
    rpy_t = np.array([0.0, 0.0, 0.3])  # Roll, pitch-up, turn rates
    H_inv = euler_kinematics(rpy, inverse=True)  # rpy_t -> w_B
    w_B_expected = H_inv @ rpy_t

    # Second, convert the quaternion derivative to angular momentum
    # See notes above on this conversion
    w_B_out = 2 * (att.inv().mul(x_t.att, normalize=False)).array[1:]
    assert np.allclose(w_B_out, w_B_expected, atol=1e-4)

    # Turn coordination when flight path angle is zero
    # This verifies equation 3.6-7 in Lewis, Johnson, Stevens
    phi = rpy[0]
    H = euler_kinematics(rpy)  # w_B -> rpy_t
    rpy_t = H @ w_B_out
    psi_t = rpy_t[2]  # Turn rate

    G = psi_t * vt / g0
    tph1 = np.tan(phi)
    tph2 = G * np.cos(beta) / (np.cos(alpha) - G * np.sin(alpha) * np.sin(beta))
    assert np.allclose(tph1, tph2)


def test_trim(trim_cases):
    """Trim conditions (Sec. 3.6-3 in Stevens, Lewis, Johnson)"""

    tolerances = {
        "alpha": 1e-3,
        "beta": 1e-3,
        "throttle": 1e-3,
        "elevator": 1e-2,
        "aileron": 1e-2,
        "rudder": 1e-1,
    }

    for case in trim_cases:
        print(case["name"])
        print(case["description"])

        f16 = SubsonicF16(xcg=case["xcg"])
        result = f16.trim(**case["condition"])

        # Check that the residuals (v_B and w_B) are small
        assert np.linalg.norm(result.residuals) < 1e-8

        # Check that the trim point matches Table 3.6-3
        for field, atol in tolerances.items():
            expected = float(case["variables"].get(field))
            actual = getattr(result.variables, field)
            close = np.allclose(expected, actual, atol=atol)
            print("\t", field, close)
            assert close


def test_linearization(trim_cases):
    """Linearization about a trim point (Sec. 3.7 in Stevens, Lewis, Johnson)"""
    case = next(case for case in trim_cases if case["name"] == "pull_up")
    print(case)

    f16 = SubsonicF16(xcg=case["xcg"])
    result = f16.trim(**case["condition"])

    x0 = StabilityState.from_full_state(result.state)
    u0 = StabilityInput.from_full_input(result.inputs)

    x0_flat, unravel_x = arc.tree.ravel(x0)
    u0_flat, unravel_u = arc.tree.ravel(u0)

    def unravel_stab(x_flat, u_flat):
        x_stab = unravel_x(x_flat)
        u_stab = unravel_u(u_flat)
        x_full = x_stab.as_full_state()
        u_full = u_stab.as_full_input()
        return x_full, u_full

    def dynamics(t, x_flat, u_flat):
        x_full, u_full = unravel_stab(x_flat, u_flat)
        x_dot_full = f16.dynamics(t, x_full, u_full)

        x_dot = StabilityState.from_full_derivative(x_full, x_dot_full)
        x_dot_flat, _ = arc.tree.ravel(x_dot)

        return x_dot_flat

    def output(t, x_flat, u_flat):
        # Output normal acceleration, pitch rate, and angle of attack
        x_full, u_full = unravel_stab(x_flat, u_flat)

        x_stab = unravel_x(x_flat)
        q = x_stab.lon.q
        alpha = np.rad2deg(x_stab.lon.alpha)  # Angle of attack [deg]

        # Normal acceleration [g's]
        F_net_B, _ = f16.net_forces(t, x_full, u_full)
        F_grav_B = f16.calc_gravity(x_full)
        a_n = -(F_net_B[2] - F_grav_B[2]) / (f16.m * GRAV_FTS2)

        return np.hstack([a_n, q, alpha])

    x0_dot_flat = dynamics(0.0, x0_flat, u0_flat)
    x0_dot_expected = StabilityState(
        lon=LongitudinalState(
            vt=0.0,
            alpha=0.0,
            theta=case["condition"]["pitch_rate"],
            q=0.0,
            eng=f16.engine.State(0.0),
        ),
        lat=LateralState(
            beta=0.0,
            phi=0.0,
            p=0.0,
            r=0.0,
        ),
    )
    assert np.allclose(
        x0_dot_flat,
        arc.tree.ravel(x0_dot_expected)[0],
        atol=1e-3,
    )

    y0 = output(0.0, x0_flat, u0_flat)
    y0_expected = np.array([5.4267, 0.3, 17.2208])
    assert np.allclose(y0, y0_expected, atol=1e-3)

    A, B = arc.jac(dynamics, argnums=(1, 2))(0.0, x0_flat, u0_flat)
    C, D = arc.jac(output, argnums=(1, 2))(0.0, x0_flat, u0_flat)

    A_lon_ex = np.array(
        [
            [-0.127, -235, -32.2, -9.51, 0.314],
            [-7e-4, -0.969, 0, 0.908, -2e-4],
            [0, 0, 0, 1, 0],
            [9e-4, -4.56, 0, -1.58, 0],
            [0, 0, 0, 0, -5],
        ]
    )
    npt.assert_allclose(A[:5, :5], A_lon_ex, rtol=1e-2, atol=1e-3)

    A_lat_ex = np.array(
        [
            [-0.322, 0.0612, 0.298, -0.948],
            [0, 0.093, 1.0, 0.310],
            [-62.5, 0, -3.0, 1.99],
            [7.67, 0, -0.262, -0.629],
        ]
    )
    npt.assert_allclose(A[5:, 5:], A_lat_ex, rtol=0.05, atol=1e-3)

    A_mix_ex = np.array(
        [
            [-0.0028, 0.00126, 5e-5, 2e-4],
            [1.5e-5, 0, -4e-5, -1e-5],
            [0, 0, 0, 0],
            [9.2e-5, 0, 0, -0.00287],
            [0, 0, 0, 0],
        ]
    )
    npt.assert_allclose(A[:5, 5:], A_mix_ex, rtol=0.1, atol=1e-2)

    A_mix_ex = np.array(
        [
            [1e-8, 2e-5, 3e-6, 8e-7, -3e-8],
            [0, 0, 0, 0, 0],
            [-3e-7, -0.00248, 0, 3e-4, 0],
            [-3e-6, -0.00188, 0, 0.00254, 0],
        ]
    )
    npt.assert_allclose(A[5:, :5], A_mix_ex, rtol=0.1, atol=1e-2)

    B_ex = np.array(
        [
            [0, -0.244, 6e-6, 2e-5],
            [0, -0.00209, 0, 0],
            [0, 0, 0, 0],
            [0, -0.199, 0, 0],
            [1087, 0, 0, 0],
            [0, 2e-8, 3e-4, 8e-4],
            [0, 0, 0, 0],
            [0, 0, -0.645, 0.126],
            [0, 0, -0.018, -0.0657],
        ]
    )
    npt.assert_allclose(B, B_ex, rtol=1e-3, atol=1e-3)

    C_ex = np.array(
        [
            [0.0208, 15.2, 0, 1.45, 0, -4.5e-4, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0, 0],
            [0, 57.3, 0, 0, 0, 0, 0, 0, 0],
        ]
    )
    npt.assert_allclose(C, C_ex, rtol=1e-2, atol=1e-3)

    D_ex = np.array(
        [
            [0, 0.0333, 0, 0],
            [0, 0, 0, 0],
            [0, 0, 0, 0],
        ]
    )
    npt.assert_allclose(D, D_ex, rtol=1e-3, atol=1e-3)


def test_lon_stability():
    """Test case for longitudinal stability"""
    # Steady pitch-up
    model = SubsonicF16(xcg=0.3)
    result = model.trim(vt=502, pitch_rate=0.0)

    x0 = LongitudinalState.from_full_state(result.state)
    x0_lat = LateralState.from_full_state(result.state)
    u0 = LongitudinalInput.from_full_input(result.inputs)

    x0_flat, unravel_x = arc.tree.ravel(x0)
    u0_flat, unravel_u = arc.tree.ravel(u0)

    def unravel_stab(x_flat, u_flat):
        x_stab = StabilityState(lon=unravel_x(x_flat), lat=x0_lat)
        u_stab = unravel_u(u_flat)
        x_full = x_stab.as_full_state()
        u_full = u_stab.as_full_input()
        return x_full, u_full

    def dynamics(t, x_flat, u_flat):
        x_full, u_full = unravel_stab(x_flat, u_flat)
        x_dot_full = model.dynamics(t, x_full, u_full)

        x_dot = StabilityState.from_full_derivative(x_full, x_dot_full)
        x_dot_flat, _ = arc.tree.ravel(x_dot.lon)
        return x_dot_flat

    A = arc.jac(dynamics, argnums=(1,))(0.0, x0_flat, u0_flat)
    evals, evecs = np.linalg.eig(A)

    phugoid_idx = np.argmax(evals.real)
    short_idx = np.argmax(evals.imag)

    # Stevens, Lewis, Johnson values:
    slj_phugoid = -0.0087 + 0.0739j
    slj_short = -1.2039 + 1.4922j

    print(
        f"Phugoid mode:\t\t{evals[phugoid_idx].real:.4f} ± "
        f"{evals[phugoid_idx].imag:.4f}j"
    )
    print(
        f"Short period mode:\t{evals[short_idx].real:.4f} ± "
        f"{evals[short_idx].imag:.4f}j"
    )

    assert np.allclose(evals[phugoid_idx], slj_phugoid, rtol=1e-3)
    assert np.allclose(evals[short_idx], slj_short, rtol=1e-3)

    v_phugoid = evecs[:, phugoid_idx]
    if np.any(v_phugoid.imag < 0):
        v_phugoid = v_phugoid.conj()  # Make imaginary part positive for comparison
    npt.assert_allclose(
        v_phugoid,
        np.array([-1.0, 9.6e-5 + 5e-7j, 3.8e-4 + 2.3e-3j, -1.7e-4 + 8.4e-6j, 0]),
        rtol=1e-2,
        atol=1e-5,
    )

    v_short = evecs[:, short_idx]
    if np.any(v_short.imag < 0):
        v_short = v_short.conj()  # Make imaginary part positive for comparison
    npt.assert_allclose(
        v_short,
        np.array([1.0, 0.09 + 0.017j, 0.059 + 0.054j, 0.0092 - 0.15j, 0]),
        rtol=5e-2,
        atol=1e-5,
    )


def test_lat_stability():
    """Test case for lateral-directional stability"""
    # Steady pitch-up
    model = SubsonicF16(xcg=0.3)
    result = model.trim(vt=502, pitch_rate=0.0)

    x0 = LateralState.from_full_state(result.state)
    x0_lon = LongitudinalState.from_full_state(result.state)
    u0 = LateralInput.from_full_input(result.inputs)

    x0_flat, unravel_x = arc.tree.ravel(x0)
    u0_flat, unravel_u = arc.tree.ravel(u0)

    def unravel_stab(x_flat, u_flat):
        x_stab = StabilityState(lon=x0_lon, lat=unravel_x(x_flat))
        u_stab = unravel_u(u_flat)
        x_full = x_stab.as_full_state()
        u_full = u_stab.as_full_input()
        return x_full, u_full

    def dynamics(t, x_flat, u_flat):
        x_full, u_full = unravel_stab(x_flat, u_flat)
        x_dot_full = model.dynamics(t, x_full, u_full)

        x_dot = StabilityState.from_full_derivative(x_full, x_dot_full)
        x_dot_flat, _ = arc.tree.ravel(x_dot.lat)

        return x_dot_flat

    A = arc.jac(dynamics, argnums=(1))(0.0, x0_flat, u0_flat)

    # Example 3.8-1 in Stevens, Lewis, Johnson
    A_ex = np.array(
        [
            [-0.32200, 0.064032, 0.038904, -0.99156],
            [0.0, 0.0, 1.0, 0.039385],
            [-30.919, 0.0, -3.6730, 0.67425],
            [9.4724, 0, -0.026358, -0.49849],
        ]
    )
    npt.assert_allclose(A, A_ex, rtol=1e-3, atol=1e-4)

    evals, evecs = np.linalg.eig(A)

    dutch_idx = np.where(evals.imag != 0)[0][0]
    spiral_idx = np.argmax(evals.real)
    roll_idx = np.argmin(evals.real)

    # Stevens, Lewis, Johnson values:
    slj_dutch = -0.4399 + 3.220j
    slj_spiral = -0.0128
    slj_roll = -3.601

    print(
        f"Dutch roll mode:\t{evals[dutch_idx].real:.4f} ± {evals[dutch_idx].imag:.4f}j"
    )
    print(f"Spiral mode:\t\t{evals[spiral_idx].real:.4f}")
    print(f"Roll subsidence mode:\t{evals[roll_idx].real:.4f}")

    assert np.allclose(evals[dutch_idx], slj_dutch, rtol=1e-3)
    assert np.allclose(evals[spiral_idx], slj_spiral, rtol=5e-3)
    assert np.allclose(evals[roll_idx], slj_roll, rtol=1e-3)

    v_spiral = evecs[:, spiral_idx].real
    npt.assert_allclose(
        v_spiral,
        np.array([0.0032, 1, -0.015, 0.063]),
        rtol=2e-2,
    )

    v_roll = evecs[:, roll_idx].real
    npt.assert_allclose(
        v_roll,
        np.array([0.002, 0.28, -1, -0.015]),
        rtol=1e-1,
    )

    v_dutch = evecs[:, dutch_idx]
    if np.any(v_dutch.imag < 0):
        v_dutch = v_dutch.conj()  # Make imaginary part positive for comparison
    npt.assert_allclose(
        v_dutch,
        np.array([-0.11 + 0.097j, -0.037 + 0.3j, 1.0, -0.29 - 0.33j]),
        rtol=2e-1,
        atol=2e-1,
    )



import abc
from typing import Tuple

import numpy as np

from archimedes import struct, StructConfig, UnionConfig
from archimedes._core.utils import find_equal


__all__ = [
    "AtmosphereModel",
    "ConstantAtmosphere",
    "ConstantAtmosphereConfig",
    "StandardAtmosphere1976",
    "StandardAtmosphere1976Config",
    "AtmosphereConfig",
]


@struct
class AtmosphereModel(metaclass=abc.ABCMeta):
    Rs: float = 287.05287  # Specific gas constant for air [J/(kg·K)]
    gamma: float = 1.4  # Adiabatic index for air [-]

    @abc.abstractmethod
    def calc_p(self, alt: float) -> float:
        """Compute pressure at given altitude."""

    @abc.abstractmethod
    def calc_T(self, alt: float) -> float:
        """Compute pressure at given altitude."""

    def __call__(self, Vt: float, alt: float) -> Tuple[float, float]:
        """Compute Mach number and dynamic pressure at given altitude and velocity."""
        p = self.calc_p(alt)
        T = self.calc_T(alt)
        rho = p / (self.Rs * T)
        amach = Vt / np.sqrt(self.gamma * self.Rs * T)  # Adiabatic Mach number
        qbar = 0.5 * rho * Vt**2
        return amach, qbar


class AtmosphereConfigBase(StructConfig):
    Rs: float = 287.05287  # Specific gas constant for air [J/(kg·K)]
    gamma: float = 1.4  # Adiabatic index for air [-]


@struct
class ConstantAtmosphere(AtmosphereModel):
    """Constant atmosphere model"""

    # Defaults based on US Standard Atmosphere, 1976: 20km altitude
    p: float = 5474.89  # Pressure [Pa]
    T: float = 216.65  # Temperature [K]

    def calc_p(self, alt: float) -> float:
        """Return constant pressure"""
        return self.p

    def calc_T(self, alt: float) -> float:
        """Return constant temperature"""
        return self.T


class ConstantAtmosphereConfig(AtmosphereConfigBase, type="constant"):
    p: float = 5474.89  # Pressure [Pa]
    T: float = 216.65  # Temperature [K]

    def build(self) -> ConstantAtmosphere:
        return ConstantAtmosphere(
            Rs=self.Rs,
            gamma=self.gamma,
            p=self.p,
            T=self.T,
        )


# Altitude [m]
h_USSA1976 = np.array([0, 11000, 20000, 32000, 47000, 51000, 71000, 84852])
# Temperature [K]
T_USSA1976 = np.array([288.15, 216.65, 216.65, 228.65, 270.65, 270.65, 214.65, 186.95])
# Pressure [Pa]
p_USSA1976 = np.array([101325, 22632.06, 5474.89, 868.02, 110.91, 66.94, 3.96, 0.3734])
# Temperature lapse rate [K/m]
L_USSA1976 = np.array([-0.0065, 0, 0.001, 0.0028, 0, 0.0028, 0, 0])


@struct
class StandardAtmosphere1976(AtmosphereModel):
    """U.S. Standard Atmosphere, 1976"""

    g0: float = 9.80665  # Gravity constant m/s^2

    def calc_p(self, alt: float) -> float:
        alt = np.fmax(0, alt)
        h1 = find_equal(alt, h_USSA1976, h_USSA1976)
        T1 = find_equal(alt, h_USSA1976, T_USSA1976)
        p1 = find_equal(alt, h_USSA1976, p_USSA1976)
        L = find_equal(alt, h_USSA1976, L_USSA1976)

        return np.where(
            L == 0,
            p1 * np.exp(-self.g0 * (alt - h1) / (self.Rs * T1)),
            p1 * (T1 / (T1 + L * (alt - h1))) ** (self.g0 / (self.Rs * L)),
        )

    def calc_T(self, alt: float) -> float:
        return np.interp(alt, h_USSA1976, T_USSA1976)


class StandardAtmosphere1976Config(AtmosphereConfigBase, type="ussa1976"):
    g0: float = 9.80665  # Gravity constant m/s^2

    def build(self) -> StandardAtmosphere1976:
        return StandardAtmosphere1976(
            g0=self.g0,
            Rs=self.Rs,
            gamma=self.gamma,
        )


AtmosphereConfig = UnionConfig[
    ConstantAtmosphereConfig,
    StandardAtmosphere1976Config,
]



import abc
import os
import re

import jinja2

__all__ = ["_render_template"]

DEFAULT_TEMPLATE_PATH = os.path.join(
    os.path.dirname(__file__),
    "_templates",
)


def _extract_protected_regions(file_path):
    """
    Extract protected regions from an existing file.

    Args:
        file_path: Path to the file with protected regions

    Returns:
        Dictionary mapping region names to their content
    """
    if not os.path.exists(file_path):
        return {}

    with open(file_path, "r") as f:
        content = f.read()

    protected_regions = {}
    pattern = r"// PROTECTED-REGION-START: (\w+)(.*?)// PROTECTED-REGION-END"
    matches = re.finditer(pattern, content, re.DOTALL)

    for match in matches:
        region_name = match.group(1)
        region_content = match.group(2).strip()
        protected_regions[region_name] = region_content

    return protected_regions


class RendererBase(metaclass=abc.ABCMeta):
    def __init__(self, template_path=None):
        if template_path is None:
            template_path = os.path.join(
                DEFAULT_TEMPLATE_PATH, self.default_template_name
            )
        self.template_path = template_path

    @property
    def default_template_name(self):
        """Default template name for this renderer."""
        raise NotImplementedError("Renderer must define a default template name.")

    def __call__(self, context, output_path):
        """
        Render a C application from a Jinja2 template.

        Args:
            context: Dictionary with template variables
            output_path: Path where the generated code will be written
        """
        template_dir = os.path.dirname(self.template_path)
        template_name = os.path.basename(self.template_path)

        context["app_name"] = os.path.basename(output_path)

        # Extract existing protected regions if the file exists
        protected_regions = {}
        if os.path.exists(output_path):
            protected_regions = _extract_protected_regions(output_path)

        # Add protected regions to the context
        context["protected_regions"] = protected_regions

        # Set up Jinja environment
        env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(template_dir or "."),
            trim_blocks=True,
            lstrip_blocks=True,
            # autoescape=True,
            autoescape=jinja2.select_autoescape(
                enabled_extensions=("html", "xml", "htm"),  # Autoescape these
                disabled_extensions=("j2", "c", "h", "cpp"),  # Don't autoescape these
                default=False,
            ),
        )
        # env.filters['escape'] = c_code_escape  # Override default escaper

        # Load template
        template = env.get_template(template_name)

        # Render template with context
        rendered_code = template.render(**context)

        # Create directory if it doesn't exist
        os.makedirs(os.path.dirname(os.path.abspath(output_path)), exist_ok=True)

        # Write output file
        with open(output_path, "w") as f:
            f.write(rendered_code)


class APIRenderer(RendererBase):
    @property
    def default_template_name(self):
        return "c_api.c.j2"


class APIHeaderRenderer(APIRenderer):
    @property
    def default_template_name(self):
        return "c_api.h.j2"


class CApplicationRenderer(RendererBase):
    @property
    def default_template_name(self):
        return "c_application.j2"


class ArduinoRenderer(RendererBase):
    @property
    def default_template_name(self):
        return "arduino.j2"


_builtin_templates = {
    "c_app": CApplicationRenderer,
    "api": APIRenderer,
    "api_header": APIHeaderRenderer,
    "arduino": ArduinoRenderer,
}


def _render_template(
    template: str | type[RendererBase],
    context: dict,
    template_path: str | None = None,
    output_path: str | None = None,
) -> RendererBase:
    """
    Render a template with the given context and save it to the specified path.

    Args:
        template: Name of the template to render, or a RendererBase instance
        context: Dictionary with template variables
        output_path: Path where the generated code will be written
        template_path: Path to the Jinja2 template file
    """
    if isinstance(template, str):
        if template not in _builtin_templates:
            raise ValueError(f"Template '{template}' not found.")

        renderer = _builtin_templates[template](template_path)

    else:
        try:
            # Will also raise a TypeError if template is not a class
            if issubclass(template, RendererBase):
                renderer = template(template_path)
            else:
                raise TypeError
        except TypeError:
            raise ValueError("Template must be a string or RendererBase class.")

    renderer(context, output_path)

    return renderer



# ruff: noqa: N802, N803, N806, E741

# Mathematical correctness tests for Extended Kalman Filter
import numpy as np
import pytest
from scipy.linalg import cholesky

from archimedes.observers import (
    ExtendedKalmanFilter,
    UnscentedKalmanFilter,
)


def get_kf(filter_name):
    return {
        "ekf": ExtendedKalmanFilter,
        "ukf": UnscentedKalmanFilter,
    }[filter_name]


@pytest.mark.parametrize("filter_name", ["ekf", "ukf"])
def test_linear_consistency(filter_name):
    """For linear systems, nonlinear filters should match linear KF."""
    KalmanFilter = get_kf(filter_name)

    # Define a simple linear system
    dt = 0.1
    F = np.array([[1.0, dt], [0.0, 1.0]])  # Constant velocity
    H = np.array([[1.0, 0.0]])  # Position measurement
    Q = np.array([[0.1, 0.0], [0.0, 0.1]])
    R = np.array([[0.5]])

    # Linear dynamics and observation functions
    def f_linear(t, x):
        return F @ x

    def h_linear(t, x):
        return H @ x

    # Initial conditions
    x0 = np.array([1.0, 0.5])
    P0 = np.eye(2) * 0.1
    y = np.array([1.1])  # Measurement

    kf = KalmanFilter(f_linear, h_linear, Q, R)

    # KF step
    x_kf, P_kf, e_kf = kf.step(0.0, x0, y, P0)

    # Manual linear KF step for comparison
    # Predict
    x_pred = F @ x0
    P_pred = F @ P0 @ F.T + Q

    # Update
    y_pred = H @ x_pred
    innovation = y - y_pred
    S = H @ P_pred @ H.T + R
    K = P_pred @ H.T @ np.linalg.inv(S)
    x_linear = x_pred + K @ innovation
    P_linear = (np.eye(2) - K @ H) @ P_pred

    # Results should be very close (UKF should be exact for linear systems)
    np.testing.assert_allclose(x_kf, x_linear, rtol=1e-10)
    np.testing.assert_allclose(P_kf, P_linear, rtol=1e-10)
    np.testing.assert_allclose(e_kf, innovation, rtol=1e-10)


@pytest.mark.parametrize("filter_name", ["ekf", "ukf"])
def test_symmetry(filter_name):
    """Test that the filter respects symmetry properties of the problem."""
    KalmanFilter = get_kf(filter_name)

    # Symmetric system
    def f_symmetric(t, x):
        return np.array([0.9 * x[0], 0.9 * x[1]], like=x)  # Symmetric dynamics

    def h_symmetric(t, x):
        return np.array(
            [x[0] ** 2 + x[1] ** 2], like=x
        )  # Rotationally symmetric measurement

    Q = np.eye(2) * 0.01
    R = np.array([[0.1]])

    kf = KalmanFilter(
        dyn=f_symmetric,
        obs=h_symmetric,
        Q=Q,
        R=R,
    )

    # Symmetric initial conditions
    x1 = np.array([1.0, 0.0])
    x2 = np.array([0.0, 1.0])  # 90-degree rotation
    P = np.eye(2) * 0.1  # Symmetric covariance

    y1 = np.array([1.0])  # Same measurement for both (due to symmetry)
    y2 = np.array([1.0])

    # UKF steps
    x1_post, P1_post, _ = kf.step(0.0, x1, y1, P)
    x2_post, P2_post, _ = kf.step(0.0, x2, y2, P)

    # Posterior states should have same magnitude
    np.testing.assert_allclose(
        np.linalg.norm(x1_post), np.linalg.norm(x2_post), rtol=1e-10
    )

    # Posterior covariances should be related by 90° rotation
    # P2 should equal R90^T * P1 * R90 where R90 is 90° rotation matrix
    R90 = np.array([[0, 1], [-1, 0]])  # 90° counter-clockwise rotation
    P1_rotated = R90.T @ P1_post @ R90
    np.testing.assert_allclose(P2_post, P1_rotated, rtol=1e-10)

    print(
        f"State magnitudes: {np.linalg.norm(x1_post):.10f} == "
        f"{np.linalg.norm(x2_post):.10f}"
    )
    print(
        "Covariance rotation relationship verified with precision: "
        f"{np.max(np.abs(P2_post - P1_rotated)):.2e}"
    )


class TestKFClasses:
    @pytest.mark.parametrize("filter_name", ["ekf", "ukf"])
    def test_class_properties(self, filter_name, simple_nonlinear_system):
        KalmanFilter = get_kf(filter_name)

        kf = KalmanFilter(
            dyn=simple_nonlinear_system["f"],
            obs=simple_nonlinear_system["h"],
            Q=simple_nonlinear_system["Q"],
            R=simple_nonlinear_system["R"],
        )

        nx = simple_nonlinear_system["Q"].shape[0]
        ny = simple_nonlinear_system["R"].shape[0]

        assert kf.nx == nx, f"{filter_name} nx property mismatch"
        assert kf.ny == ny, f"{filter_name} ny property mismatch"


class TestEKF:
    """Test mathematical properties and correctness of EKF implementation."""

    def test_measurement_update_properties(self):
        """Test mathematical properties of the measurement update step."""

        def h(t, x):
            return np.array([x[0] + 0.1 * x[1] ** 2], like=x)  # Nonlinear measurement

        R = np.array([[0.1]])

        ekf = ExtendedKalmanFilter(None, h, None, R)

        # State and covariance BEFORE measurement update
        x_prior = np.array([1.0, 0.5])
        P_prior = np.array([[0.1, 0.02], [0.02, 0.08]])
        y = np.array([1.02])

        # Apply ONLY the measurement update
        x_post, P_post, innovation = ekf.correct(0.0, x_prior, y, P_prior)

        # Test 1: Posterior covariance should be smaller than prior
        assert np.trace(P_post) < np.trace(P_prior), (
            "Measurement should reduce uncertainty"
        )

        # Test 2: Innovation should be consistent with prediction
        y_pred = h(0.0, x_prior)
        expected_innovation = y - y_pred
        np.testing.assert_allclose(innovation, expected_innovation, rtol=1e-12)

        # Test 3: Information matrix should increase
        info_prior = np.linalg.inv(P_prior)
        info_post = np.linalg.inv(P_post)

        # Information should not decrease (Loewner ordering)
        diff = info_post - info_prior
        eigenvals = np.linalg.eigvals(diff)
        assert np.all(eigenvals >= -1e-10), "Information matrix should not decrease"


class TestUKF:
    """Test mathematical properties and correctness of UKF implementation."""

    def test_unscented_transform(self):
        # Define test mean and covariance
        x_mean = np.array([1.0, -0.5])
        P = np.array([[0.1, 0.02], [0.02, 0.08]])

        # Test polynomial functions
        def linear_func(x):
            """Linear function: should be exact"""
            return np.array([2 * x[0] + 3 * x[1], -x[0] + 4 * x[1]])

        def quadratic_func(x):
            """Quadratic function: should be exact for unscented transform"""
            return np.array([x[0] ** 2, x[1] ** 2, x[0] * x[1]])

        # Generate sigma points using Julier method (matches our UKF)
        L = len(x_mean)
        kappa = 0.0

        # FIXED: Use scipy.linalg.cholesky and proper indexing
        U = cholesky((L + kappa) * P)

        sigma_points = []
        sigma_points.append(x_mean)

        for j in range(L):
            # FIXED: Use U[j] (j-th row) instead of A[:, j] (j-th column)
            sigma_points.append(x_mean + U[j])
            sigma_points.append(x_mean - U[j])

        # FIXED: Use proper Julier weights instead of equal weights
        w = np.full(2 * L + 1, 0.5 / (L + kappa))
        w[0] = kappa / (L + kappa)

        # Test linear function
        y_points_linear = np.array([linear_func(sp) for sp in sigma_points])
        y_mean_ut = np.sum(w[:, np.newaxis] * y_points_linear, axis=0)

        # Analytical result for linear function
        A_matrix = np.array([[2, 3], [-1, 4]])
        y_mean_analytical = A_matrix @ x_mean

        np.testing.assert_allclose(y_mean_ut, y_mean_analytical, rtol=1e-12)

        # Test quadratic function
        y_points_quad = np.array([quadratic_func(sp) for sp in sigma_points])
        y_mean_ut_quad = np.sum(w[:, np.newaxis] * y_points_quad, axis=0)

        # Analytical moments for quadratic function
        expected_quad = np.array(
            [
                P[0, 0] + x_mean[0] ** 2,  # E[x₁²]
                P[1, 1] + x_mean[1] ** 2,  # E[x₂²]
                P[0, 1] + x_mean[0] * x_mean[1],  # E[x₁x₂]
            ]
        )

        np.testing.assert_allclose(y_mean_ut_quad, expected_quad, rtol=1e-10)

    def test_measurement_update_information_increase(self):
        """Test that measurement updates increase information (decrease uncertainty)."""

        def f(t, x):
            return x  # Identity dynamics for simplicity

        def h(t, x):
            return np.array([x[0] + 0.1 * x[1] ** 2, x[1]])  # Nonlinear measurement

        Q = np.eye(2) * 0.001  # Small process noise
        R = np.eye(2) * 0.1

        ukf = UnscentedKalmanFilter(f, h, Q, R)

        x_prior = np.array([1.0, 0.5])
        P_prior = np.array([[0.2, 0.05], [0.05, 0.15]])
        y = np.array([1.02, 0.48])

        x_post, P_post, _ = ukf.step(0.0, x_prior, y, P_prior)

        # Test: Posterior uncertainty should be smaller
        trace_prior = np.trace(P_prior + Q)  # Include process noise
        trace_post = np.trace(P_post)

        assert trace_post < trace_prior, "Measurement should reduce uncertainty"

        # Test: Determinant should decrease (volume of uncertainty ellipsoid)
        det_prior = np.linalg.det(P_prior + Q)
        det_post = np.linalg.det(P_post)

        assert det_post < det_prior, "Measurement should reduce uncertainty volume"

    def test_degenerate_covariance_handling(self):
        """Test UKF behavior with near-singular covariance matrices."""

        def f(t, x):
            return np.array([0.99 * x[0], 0.99 * x[1]])

        def h(t, x):
            return np.array([x[0]])  # Only observe first state

        Q = np.eye(2) * 1e-6  # Very small process noise
        R = np.array([[0.01]])

        ukf = UnscentedKalmanFilter(f, h, Q, R)

        # Start with well-conditioned covariance
        x = np.array([1.0, 1.0])
        P = np.eye(2) * 0.1

        # Run many steps with only partial observations
        # This should make P become nearly singular in unobserved direction
        for i in range(50):
            y = np.array([1.0 + 0.01 * np.random.randn()])
            x, P, _ = ukf.step(i * 0.1, x, y, P)

            # Should not crash even if P becomes ill-conditioned
            assert np.all(np.isfinite(x)), f"State became non-finite at step {i}"
            assert np.all(np.isfinite(P)), f"Covariance became non-finite at step {i}"

            # Eigenvalues should remain non-negative
            eigenvals = np.linalg.eigvals(P)
            assert np.all(eigenvals >= -1e-12), (
                f"Negative eigenvalues at step {i}: {eigenvals}"
            )



"""Test N-dimensional interpolation"""

# ruff: noqa: N802
# ruff: noqa: N803
# ruff: noqa: N806

import numpy as np
import pytest
from scipy.interpolate import RectBivariateSpline

import archimedes as arc
from archimedes._core import SymbolicArray, compile, sym


class TestNumPyInterp:
    def test_interp1d(self):
        xp = np.linspace(1, 6, 6)
        fp = np.array([-1, -1, -2, -3, 0, 2])

        @compile(kind="MX")
        def f(x):
            return np.interp(x, xp, fp)

        # Test evaluating symbolically (note must be "MX")
        x = sym("x", kind="MX")
        y = np.interp(x, xp, fp)
        assert isinstance(y, SymbolicArray)
        assert y.shape == ()
        assert y.dtype == x.dtype

        # Test evaluating numerically
        x = 2.5
        assert np.isclose(f(x), np.interp(x, xp, fp))

        # Above and below the range
        assert f(0) == -1
        assert f(7) == 2

        # Test array evaluation
        x = np.array([2.5, 3.5])
        assert np.allclose(f(x), np.interp(x, xp, fp))

    def test_error_handling(self):
        xp = np.linspace(1, 6, 6)
        fp = np.array([-1, -1, -2, -3, 0, 2])

        # Test with SX (should through an error)
        with pytest.raises(ValueError, match=r".*MX.*"):
            x = sym("x", kind="SX")
            np.interp(x, xp, fp)

        x = sym("x", kind="MX")
        with pytest.raises(ValueError, match=r"xp must be 1-dimensional.*"):
            np.interp(x, xp[:, None], fp)

        with pytest.raises(ValueError, match=r"fp must be 1-dimensional.*"):
            np.interp(x, xp, fp[:, None])

        with pytest.raises(ValueError, match=r".*must be NumPy arrays.*"):
            np.interp(x, arc.sym_like(xp, "xp"), fp)

        with pytest.raises(ValueError, match=r".*must be scalars.*"):
            np.interp(x, xp, fp, left=xp)


class TestNDInterp:
    def test_interp1d(self):
        xp = np.linspace(1, 6, 6)
        fp = np.array([-1, -1, -2, -3, 0, 2])

        f_lut = arc.interpolant([xp], fp, arg_names=("x",), ret_name="y", name="f")

        # Check names
        assert f_lut.arg_names == [
            "x",
        ]
        assert f_lut.return_names == [
            "y",
        ]
        assert f_lut.name == "f"

        # Test evaluating symbolically (note must be "MX")
        x = sym("x", kind="MX")
        y = f_lut(x)
        assert isinstance(y, SymbolicArray)
        assert y.shape == ()
        assert y.dtype == x.dtype

        # Test evaluating numerically
        x = 2.5
        assert np.isclose(f_lut(x), np.interp(x, xp, fp))

        # Test array evaluation
        x = np.array([2.5, 3.5, 4.5])
        y = f_lut(x)
        assert y.shape == (3,)
        assert np.allclose(y, np.interp(x, xp, fp))

    def test_interp2d(self):
        # https://web.casadi.org/docs/#using-lookup-tables
        xgrid = np.linspace(-5, 5, 11)
        ygrid = np.linspace(-4, 4, 9)
        X, Y = np.meshgrid(xgrid, ygrid, indexing="ij")
        R = np.sqrt(5 * X**2 + Y**2) + 1
        Z = np.sin(R) / R
        lut = arc.interpolant(
            [xgrid, ygrid],
            Z,
            arg_names=("x", "y"),
            ret_name="z",
            name="f",
            method="bspline",
        )

        # Check names
        assert lut.arg_names == ["x", "y"]
        assert lut.return_names == [
            "z",
        ]
        assert lut.name == "f"

        # Test numeric scalar evaluation
        x, y = 0.5, 1
        z_arc = lut(x, y)
        interp = RectBivariateSpline(xgrid, ygrid, Z)
        z_scipy = interp.ev(x, y)
        assert np.allclose(z_arc, z_scipy)

        # Test symbolic array evaluation
        n = 3
        x = sym("x", shape=(n,), kind="MX")
        y = sym("y", shape=(n,), kind="MX")
        z = lut(x, y)
        assert z.shape == (n,)
        assert z.dtype == x.dtype

        # Test numeric array evaluation
        x, y = [0.5, 0.75], [1.0, 1.25]
        z_arc = lut(x, y)
        interp = RectBivariateSpline(xgrid, ygrid, Z)
        z_scipy = interp.ev(x, y)
        assert np.allclose(z_arc, z_scipy)

    def test_default_args(self):
        # https://web.casadi.org/docs/#using-lookup-tables
        xgrid = np.linspace(-5, 5, 11)
        ygrid = np.linspace(-4, 4, 9)
        X, Y = np.meshgrid(xgrid, ygrid, indexing="ij")
        R = np.sqrt(5 * X**2 + Y**2) + 1
        Z = np.sin(R) / R
        arc.interpolant(
            [xgrid, ygrid],
            Z,
        )

    def test_error_handling(self):
        xgrid = np.linspace(-5, 5, 11)
        ygrid = np.linspace(-4, 4, 9)
        X, Y = np.meshgrid(xgrid, ygrid, indexing="ij")
        R = np.sqrt(5 * X**2 + Y**2) + 1
        Z = np.sin(R) / R
        lut = arc.interpolant(
            [xgrid, ygrid],
            Z,
            arg_names=("x", "y"),
            ret_name="z",
            name="f",
            method="bspline",
        )

        # Wrong number of arguments
        with pytest.raises(TypeError, match=r".*too many positional arguments.*"):
            x = 0.5
            lut(x, x, x)

        # Too many dimensions
        with pytest.raises(ValueError, match=r".*0- or 1-dimensional.*"):
            x = np.ones((2, 2))
            lut(x, 1.0)

        # Inconsistent lengths of point arrays
        with pytest.raises(ValueError, match=r".*must have the same length.*"):
            lut(np.array([0.5, 0.75, 1.0]), np.array([1.0, 1.5]))

        # Too many dimensions for grid points
        with pytest.raises(ValueError, match=r"grid\[1\] must be 1-dimensional.*"):
            lut = arc.interpolant(
                [xgrid, ygrid[:, None]],
                Z,
            )

        # Too many dimensions for data array
        with pytest.raises(ValueError, match=r"data must be 2-dimensional.*"):
            lut = arc.interpolant(
                [xgrid, ygrid],
                Z[:, None],
            )

        # Not enough data points
        with pytest.raises(ValueError, match=r"data must have length.*"):
            lut = arc.interpolant(
                [xgrid, ygrid],
                Z[:4],
            )

        # Invalid method
        with pytest.raises(ValueError, match=r"method must be one of.*"):
            lut = arc.interpolant(
                [xgrid, ygrid],
                Z,
                method="something else",
            )

        # Not enough arg names
        with pytest.raises(ValueError, match=r"arg_names must have length 2.*"):
            lut = arc.interpolant(
                [xgrid, ygrid],
                Z,
                arg_names=["x"],
            )

        # Non-string arg names
        with pytest.raises(ValueError, match=r"arg_names must be a list of strings.*"):
            lut = arc.interpolant(
                [xgrid, ygrid],
                Z,
                arg_names=[0, 1],
            )

        # Non-string return names
        with pytest.raises(ValueError, match=r"ret_name must be a string.*"):
            lut = arc.interpolant(
                [xgrid, ygrid],
                Z,
                ret_name=0,
            )
